package org.opensourcephysics.tuleja.action;

// (c) Slavomir Tuleja, June 24, 2002
// Revised January 5-11, 2003
// Revised January 19, 2003
// Revised January 24-27, 2003
// Revised March 12, 2003

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.AffineTransform;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;

public class JAction
    extends JApplet
    implements Runnable, MouseListener, MouseMotionListener, ActionListener {

  Scale sc;
  int numOfLayers;
  double t1, t2, h1, h2;
  double deltaT;
  double m = 0.2, g = 9.8;
  double hMax, hC, tC, SMin;
  double[] h;
  double[] sABMin;
  private BufferedImage image;
  private Graphics2D g2;
  public Canvas canvas;
  Container cp;
  JPanel pnlSouth;
  JButton btnHuntPause, btnReset, btnZoomInOut, btnDrag;
  JComboBox cbZoom;
  ComboHandler comboHandler;
  Thread thread = null;
  Cursor crCrossHair, crHand, crDefault;
  boolean isNotHunting = true, cursorOverDiagram = false, isZoomed = false,
      isZoomDrag = false, drawTableRectangle = false;
  String pH1, pT1, pH2, pT2, pNumOfLayers, pAppletMode, pMovableEnds,
      pEnableHunting, pEnableZoom;
  int windowWidth = 700, windowHeight = 450;
  int diagramWidth; //width and height of spacetime diagram in pixels
  int diagramHeight = 340;
  int leftSpace = 70;
  int topSpace = 10;
  double zoomRectW, zoomRectH;
  double zoomCenterX, zoomCenterY;
  int appletMode, movableEnds, enableHunting, enableZoom;
  double zoom = 16;
  int kD;

  public void readParameters() {
    //read the parameters from html file....
    pNumOfLayers = new String(getParameter("numOfLayers"));
    pH1 = new String(getParameter("h1"));
    pT1 = new String(getParameter("t1"));
    pH2 = new String(getParameter("h2"));
    pT2 = new String(getParameter("t2"));
    pAppletMode = new String(getParameter("appletMode"));
    pMovableEnds = new String(getParameter("movableEnds"));
    pEnableHunting = new String(getParameter("enableHunting"));
    pEnableZoom = new String(getParameter("enableZoom"));
  }

  public void createGUI() {
    //creation of GUI
    //setting Java look and feel
    try {
      UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
    }
    catch (Exception e) {
      System.err.println("Couldn't use the cross-platform " + "look and feel: " +
                         e);
    }

    //buttons
    btnHuntPause = new JButton("Hunt");
    btnHuntPause.addActionListener(this);
    btnReset = new JButton("Reset");
    btnReset.addActionListener(this);
    btnZoomInOut = new JButton("Zoom in");
    btnZoomInOut.addActionListener(this);
    btnDrag = new JButton("Start zoom drag");
    btnDrag.addActionListener(this);

    String[] sa = {
        "Zoom 16 x", "Zoom 8 x", "Zoom 4 x", "Zoom 2 x"};
    cbZoom = new JComboBox(sa);
    comboHandler = new ComboHandler(this);
    cbZoom.addActionListener(comboHandler);

    //parameters evaluated
    h1 = Double.valueOf(pH1).doubleValue();
    h2 = Double.valueOf(pH2).doubleValue();
    t1 = Double.valueOf(pT1).doubleValue();
    t2 = Double.valueOf(pT2).doubleValue();
    numOfLayers = Integer.valueOf(pNumOfLayers).intValue();
    appletMode = Integer.valueOf(pAppletMode).intValue();
    movableEnds = Integer.valueOf(pMovableEnds).intValue();
    enableHunting = Integer.valueOf(pEnableHunting).intValue();
    enableZoom = Integer.valueOf(pEnableZoom).intValue();

    if (enableHunting == 0) { //if parameter pEnableHunting is 0 we don't want buttons to be visible
      btnHuntPause.setVisible(false);
    }

    if (enableZoom != 1) {
      btnZoomInOut.setVisible(false);
      cbZoom.setVisible(false);
      btnDrag.setVisible(false);
    }

   //appletMode
    if (appletMode == 0)
      diagramWidth = windowWidth - leftSpace - 20; //applet without numerical table on right
    else if (appletMode == 1)
      diagramWidth = 4 * (windowWidth - leftSpace) / 10; // applet with acceleration table
    else if (appletMode == 2)
      diagramWidth = 4 * (windowWidth - leftSpace) / 10; // applet with energy table
    else if (appletMode == 3)
      diagramWidth = 4 * (windowWidth - leftSpace) / 10; // applet with lagrangian and energy table
    else if (appletMode == 4)
      diagramWidth = 4 * (windowWidth - leftSpace) / 10; // applet with acceleration, Energy, action on segments and segment pairs and segment-pair action minima
    else if (appletMode == 5)
      diagramWidth = 4 * (windowWidth - leftSpace) / 10; // applet with action on segments and thier total (Display#5)

   //components added to cp
    pnlSouth = new JPanel();
    if (enableZoom == 1) {
      pnlSouth.setLayout(new GridLayout(1, 5));
      pnlSouth.add(btnHuntPause);
      pnlSouth.add(btnReset);
      pnlSouth.add(btnDrag);
      pnlSouth.add(btnZoomInOut);
      pnlSouth.add(cbZoom);
    }
    else {
      pnlSouth.setLayout(new GridLayout(1, 4));
      pnlSouth.add(btnDrag);
      pnlSouth.add(btnHuntPause);
      pnlSouth.add(btnReset);
      pnlSouth.add(btnZoomInOut);
    }
    pnlSouth.setBackground(Color.white);

    image = new BufferedImage(windowWidth, windowHeight,
                              BufferedImage.TYPE_INT_RGB);
    g2 = image.createGraphics();
    canvas = new Canvas(image, this);
    canvas.addMouseListener(this);
    canvas.addMouseMotionListener(this);

    cp = getContentPane();
    cp.setLayout(new BorderLayout());
    cp.add(canvas, BorderLayout.CENTER);
    cp.add(pnlSouth, BorderLayout.SOUTH);

    //cursors definitions
    crCrossHair = new Cursor(Cursor.CROSSHAIR_CURSOR);
    crHand = new Cursor(Cursor.HAND_CURSOR);
    crDefault = new Cursor(Cursor.DEFAULT_CURSOR);

    initialSetup();

  }

  public void init() {
    readParameters();
    createGUI();
  }

  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Hunt")) {
      btnHuntPause.setText("Pause");
      isNotHunting = false;
      btnReset.setEnabled(false);
      btnZoomInOut.setEnabled(false);
      btnDrag.setEnabled(false);
      cbZoom.setEnabled(false);
      thread = new Thread(this);
      thread.start();
    }
    else if (e.getActionCommand().equals("Pause")) {
      btnHuntPause.setText("Hunt");
      if (!isZoomed) {
        btnReset.setEnabled(true);
        btnZoomInOut.setEnabled(true);
        btnDrag.setEnabled(true);
        cbZoom.setEnabled(true);
      }
      else { //if is zoomed
        btnReset.setEnabled(true);
        btnZoomInOut.setEnabled(true);
        btnDrag.setEnabled(false);
        cbZoom.setEnabled(true);
      }

      isNotHunting = true;
    }
    else if (e.getActionCommand().equals("Reset")) {
      drawTableRectangle = false;
      resetWorldline();
      canvas.repaint();
    }
    else if (e.getActionCommand().equals("Zoom in")) {
      isZoomed = true;
      btnDrag.setEnabled(false);
      //cbZoom.setEnabled(false);
      btnZoomInOut.setText("Zoom out");
      zoomScaleSetup(hMax);
    }
    else if (e.getActionCommand().equals("Zoom out")) {
      isZoomed = false;
      btnDrag.setEnabled(true);
      cbZoom.setEnabled(true);
      btnZoomInOut.setText("Zoom in");
      zoomScaleSetup(hMax);
    }
    else if (e.getActionCommand().equals("Start zoom drag")) {
      btnDrag.setText("Stop zoom drag");
      btnHuntPause.setEnabled(false);
      btnReset.setEnabled(false);
      btnZoomInOut.setEnabled(false);
      cbZoom.setEnabled(false);
      isZoomDrag = true;
      canvas.repaint();
    }
    else if (e.getActionCommand().equals("Stop zoom drag")) {
      btnDrag.setText("Start zoom drag");
      btnHuntPause.setEnabled(true);
      btnReset.setEnabled(true);
      btnZoomInOut.setEnabled(true);
      cbZoom.setEnabled(true);
      isZoomDrag = false;
      canvas.repaint();
    }

  }

  public void rescale(double h1, double h2) {
    if ( (h1 <= 50) && (h2 <= 50)) {
      double v0 = (h2 - h1 + 0.5 * g * t2 * t2) / t2;
      hMax = h1 + v0 * v0 / (2.0 * g);
      //sc=new Scale(t1,t2,0,1.2*hMax,leftSpace,leftSpace+diagramWidth,topSpace,topSpace+diagramHeight);
      zoomScaleSetup(hMax);
      zoomRectW = (t2 - t1) / zoom;
      zoomRectH = hMax / zoom;
      if (zoomCenterX < zoomRectW / 2)
        zoomCenterX = zoomRectW / 2;
      if (zoomCenterX > t2 - zoomRectW / 2)
        zoomCenterX = t2 - zoomRectW / 2;
      if (zoomCenterY < zoomRectH / 2)
        zoomCenterY = zoomRectH / 2;
      if (zoomCenterY > 1.2 * hMax - zoomRectH / 2)
        zoomCenterY = 1.2 * hMax - zoomRectH / 2;
    }
  }

  public void initialSetup() {
    deltaT = (t2 - t1) / numOfLayers;
    double v0 = (h2 - h1 + 0.5 * g * t2 * t2) / t2;
    hMax = h1 + v0 * v0 / (2.0 * g);

    //initial zoom rectangle setup
    zoomRectW = (t2 - t1) / zoom;
    zoomRectH = hMax / zoom;
    zoomCenterX = t1 + zoomRectW / 2;
    zoomCenterY = h1 + zoomRectH / 2;

    zoomScaleSetup(hMax);
    resetWorldline();

    drawBottomLayer(g2);
    SMin = action();
    canvas.repaint();
  }

  public void resetWorldline() {
    //default heights for worldline
    h = new double[numOfLayers + 1];
    for (int i = 0; i < numOfLayers + 1; i++) {
      h[i] = h1 + (h2 - h1) / (t2 - t1) * (i * deltaT);
    }
    h[0] = h1;
    h[numOfLayers] = h2;
    SMin = action();
    // strating values for sABMin[]
    if ( (appletMode == 4) || (appletMode == 5)) {
      sABMin = new double[numOfLayers - 1];
      for (int i = 0; i < numOfLayers - 1; i++) {
        sABMin[i] = sAPlusSB(i + 1);
      }
    }
  }

  public void updateZoom() {
    //called by comboHandler to change zoom value
    int index = cbZoom.getSelectedIndex();
    if (index == 0) { //Zoom 16 x
      zoom = 16;
    }
    else if (index == 1) { //Zoom 8 x
      zoom = 8;
    }
    else if (index == 2) { //Zoom 4 x
      zoom = 4;
    }
    else { //Zoom 2 x
      zoom = 2;
    }
    zoomRectW = (t2 - t1) / zoom;
    zoomRectH = hMax / zoom;
    if (zoomCenterX < zoomRectW / 2)
      zoomCenterX = zoomRectW / 2;
    if (zoomCenterX > t2 - zoomRectW / 2)
      zoomCenterX = t2 - zoomRectW / 2;
    if (zoomCenterY < zoomRectH / 2)
      zoomCenterY = zoomRectH / 2;
    if (zoomCenterY > 1.2 * hMax - zoomRectH / 2)
      zoomCenterY = 1.2 * hMax - zoomRectH / 2;
    zoomScaleSetup(hMax);
    canvas.repaint();
  }

  public void zoomScaleSetup(double hMax) {
    //scale setup
    if (!isZoomed) {
      sc = new Scale(t1, t2, 0, 1.2 * hMax, leftSpace, leftSpace + diagramWidth,
                     topSpace, topSpace + diagramHeight);
    }
    else {
      sc = new Scale(zoomCenterX - zoomRectW / 2.0,
                     zoomCenterX + zoomRectW / 2.0,
                     zoomCenterY - zoomRectH / 2.0,
                     zoomCenterY + zoomRectH / 2.0, leftSpace,
                     leftSpace + diagramWidth, topSpace,
                     topSpace + diagramHeight);
    }
    drawBottomLayer(g2);
    canvas.repaint();
  }

  public void drawBottomLayer(Graphics2D g2) {
    g2.setRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING,
                                            RenderingHints.VALUE_ANTIALIAS_ON));
    g2.setColor(Color.white);
    g2.fillRect(0, 0, windowWidth, windowHeight);

    //spacetime diagram outline
    g2.setColor(Color.black);
    g2.drawRect(leftSpace, topSpace, diagramWidth, diagramHeight);

    //vertical grid lines
    g2.setColor(new Color(200, 200, 200));
    for (int i = 1; i <= numOfLayers - 1; i++) {
      int ii = sc.xToi(t1 + i * deltaT);
      if ( (ii > leftSpace) && (ii < leftSpace + diagramWidth)) {
        g2.drawLine(ii, topSpace + diagramHeight - 1, ii, topSpace + 1);
      }
    }

    //font setup
    Font f1 = new Font("SansSerif", Font.PLAIN, 12);
    g2.setFont(f1);
    int height = g2.getFontMetrics().getHeight(); //font height

    //axes labels
    g2.setColor(Color.black);
    MyUtils.drawString("T I M E   s e c o n d s", leftSpace + diagramWidth / 2,
                       topSpace + diagramHeight + 40, 2, f1, g2);
    AffineTransform at = new AffineTransform();
    at.rotate( -Math.PI / 2.0, windowWidth / 2, windowHeight / 2);
    g2.setTransform(at);
    g2.drawString("H E I G H T   m e t e r s", 340, -105);
    at.rotate(Math.PI / 2.0, windowWidth / 2, windowHeight / 2);
    g2.setTransform(at);
  }

  public void drawTopLayer(Graphics2D g2) {
    g2.setRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING,
                                            RenderingHints.VALUE_ANTIALIAS_ON));
    int i;
    //we make sure that the worldline doesn't escape from the UNZOOMED diagram
    for (i = 1; i < numOfLayers; i++) {
      if (h[i] < 0)
        h[i] = 0;
      if (h[i] > 1.2 * hMax)
        h[i] = 1.2 * hMax;
    }

    //broken line
    g2.setColor(Color.black);
    BasicStroke bsThick, bsThin;
    bsThick = new BasicStroke(2.0f, BasicStroke.CAP_ROUND,
                              BasicStroke.JOIN_ROUND);
    bsThin = new BasicStroke(1.0f, BasicStroke.CAP_ROUND,
                             BasicStroke.JOIN_ROUND);
    //worldline
    g2.setStroke(bsThick);
    for (i = 0; i <= numOfLayers; i++) {
      if (i != numOfLayers)
        myDrawLine(t1 + i * deltaT, h[i], t1 + (i + 1) * deltaT, h[i + 1], g2);
    }
    //points of worldline
    g2.setStroke(bsThin);
    for (i = 0; i <= numOfLayers; i++) {
      int ii = sc.xToi(t1 + i * deltaT);
      int jj = sc.yToj(h[i]);
      if (intPointInside(ii, jj)) {
        if ( (movableEnds == 0) && ( (i == 0) || (i == numOfLayers))) {
          g2.setColor(Color.white);
          g2.fillOval(ii - 4, jj - 4, 8, 8);
          g2.setColor(Color.black);
          g2.drawOval(ii - 4, jj - 4, 8, 8);
        }
        else if (isZoomDrag || !isNotHunting) {
          g2.setColor(Color.white);
          g2.fillOval(ii - 4, jj - 4, 8, 8);
          g2.setColor(Color.black);
          g2.drawOval(ii - 4, jj - 4, 8, 8);
        }
        else
          g2.fillOval(ii - 4, jj - 4, 8, 8);
      }
    }

    //vertical axis ticks
    double stepH, hStart, hh;
    int exp;
    exp = Math.round( (float) Math.floor(Math.log(sc.jToy(topSpace) -
                                                  sc.jToy(topSpace +
        diagramHeight)) / Math.log(2)) - 3);
    stepH = Math.pow(2, exp);
    hStart = Math.round(sc.jToy(topSpace + diagramHeight) / stepH) * stepH;
    hh = hStart;
    do {
      int jj = sc.yToj(hh);
      if ( (jj < topSpace + diagramHeight) && (jj > topSpace))
        g2.drawLine(leftSpace, jj, leftSpace + 3, jj);
      hh += stepH;
    }
    while (hh < sc.jToy(topSpace));

    //font setup
    Font f1 = new Font("SansSerif", Font.PLAIN, 12);
    int height = g2.getFontMetrics().getHeight();

    //labels on the axes
    int numDigits;
    String s;

    //vertical axis labels
    numDigits = getNumDigits(exp);
    hh = hStart;
    do {
      s = MyUtils.roundOff(hh, numDigits);
      int jj = sc.yToj(hh);
      if ( (jj <= topSpace + diagramHeight) && (jj >= topSpace))
        MyUtils.drawString(s, leftSpace - 10, sc.yToj(hh), 1, f1, g2);
      hh += 2 * stepH;
    }
    while (hh < sc.jToy(topSpace));

    //time axis
    int stepX;
    if (numOfLayers <= 4) {
      stepX = 1;
    }
    else if (numOfLayers <= 10) {
      stepX = 2;
    }
    else {
      stepX = 4;
    }
    if (isZoomed)
      stepX = 1;

    exp = Math.round( (float) Math.floor(Math.log(sc.iTox(leftSpace +
        diagramWidth) - sc.iTox(leftSpace)) / Math.log(2)) - 1);
    numDigits = getNumDigits(exp);

    for (i = 0; i <= numOfLayers; i += stepX) {
      int ii = sc.xToi(t1 + i * deltaT);
      if ( (ii > leftSpace - 2) && (ii < leftSpace + diagramWidth + 2)) {
        s = MyUtils.roundOff(t1 + i * deltaT, numDigits);
        MyUtils.drawString(s, ii, topSpace + diagramHeight + 16, 2, f1, g2);
      }
    }

    //drawing ZOOM RED RECTANGLE
    if ( (!isZoomed) && (enableZoom == 1)) {
      g2.setColor(Color.red);
      int ii = sc.xToi(zoomCenterX - zoomRectW / 2.0);
      int jj = sc.yToj(zoomCenterY + zoomRectH / 2.0);
      int wwI = sc.xToi(zoomCenterX + zoomRectW / 2.0) -
          sc.xToi(zoomCenterX - zoomRectW / 2.0);
      int wwJ = sc.yToj(zoomCenterY - zoomRectH / 2.0) -
          sc.yToj(zoomCenterY + zoomRectH / 2.0);
      //rectangle outline
      g2.drawRect(ii, jj, wwI, wwJ);
      //rectangle center square
      if (isZoomDrag) {
        ii = sc.xToi(zoomCenterX);
        jj = sc.yToj(zoomCenterY);
        g2.fillRect(ii - 3, jj - 3, 6, 6);
      }
      g2.setColor(Color.black);
    }

    //printing S, SMin
    double S = action();
    if (SMin > S)
      SMin = S;

    MyUtils.drawString("S = ".concat(MyUtils.roundOff(S, 4)).concat(
        " joule-sec"), leftSpace - 25, topSpace + diagramHeight + 64, 3, f1, g2);
    MyUtils.drawString("S      = ".concat(MyUtils.roundOff(SMin, 4)).concat(
        " joule-sec"), leftSpace - 25, topSpace + diagramHeight + 80, 3, f1, g2);
    MyUtils.drawString("min", leftSpace + 9 - 25, topSpace + diagramHeight + 85,
                       3, new Font("SansSerif", Font.PLAIN, 9), g2);
    if (cursorOverDiagram) {
      MyUtils.drawString("time = ".concat(MyUtils.roundOff(tC, 2)).concat(
          " sec"), leftSpace + 2 * diagramWidth / 3 - 20,
                         topSpace + diagramHeight + 64, 3, f1, g2);
      MyUtils.drawString("height = ".concat(MyUtils.roundOff(hC, 2)).concat(
          " meters"), leftSpace + 2 * diagramWidth / 3 - 20,
                         topSpace + diagramHeight + 80, 3, f1, g2);
    }

    //tables
    int diagramTableSpace = 10;
    int tableWidth = windowWidth - leftSpace - diagramWidth - diagramTableSpace;
    int colWidth;
    if (appletMode == 1) { //accelerations
      colWidth = tableWidth / 4;
      MyUtils.drawString("segment",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("velocity",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("\u0394 velocity",
                         diagramTableSpace + leftSpace + diagramWidth +
                         2 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2); //unicode for delta
      MyUtils.drawString("acceleration",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);

      MyUtils.drawString("#",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("meters/sec",
                         diagramTableSpace + leftSpace + diagramWidth +
                         1 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("meters/sec",
                         diagramTableSpace + leftSpace + diagramWidth +
                         2 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2); //unicode for delta
      MyUtils.drawString("meters/sec  ",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("                     2",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height + height / 2 - 3,
                         2, new Font("SansSerif", Font.PLAIN, 9), g2);

      for (i = 1; i <= numOfLayers; i++) {
        s = String.valueOf(i);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(velocity(i), 2);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 1 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        if (i != numOfLayers) {
          s = MyUtils.roundOff(deltaVel(i), 2);
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth +
                             2 * colWidth + colWidth / 2,
                             (i + 2) * height + height / 2, 2, f1, g2);
          s = MyUtils.roundOff(acceleration(i), 1);
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth +
                             3 * colWidth + colWidth / 2,
                             (i + 2) * height + height / 2, 2, f1, g2);
        }
      }

      MyUtils.drawString("\u0394t = ".concat(MyUtils.roundOff(deltaT, 2)).
                         concat(" sec"),
                         leftSpace + diagramWidth + 2 * colWidth,
                         topSpace + diagramHeight + 40, 2, f1, g2);

    }
    else if (appletMode == 2) { //energies
      colWidth = tableWidth / 4;
      MyUtils.drawString("segment",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("T",
                         diagramTableSpace + leftSpace + diagramWidth + colWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("V",
                         diagramTableSpace + leftSpace + diagramWidth + 2 * colWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("E = T + V",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);

      MyUtils.drawString("#",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth + colWidth +
                         colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         2 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2); //unicode for delta
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);

      for (i = 1; i <= numOfLayers; i++) {
        s = String.valueOf(i);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(kinEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(potEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 2 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(kinEnergy(i) + potEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 3 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
      }

      MyUtils.drawString("\u0394t = ".concat(MyUtils.roundOff(deltaT, 2)).
                         concat(" sec"),
                         leftSpace + diagramWidth + 2 * colWidth,
                         topSpace + diagramHeight + 40, 2, f1, g2);

    }
    else if (appletMode == 3) { //lagrangians and enerigies
      colWidth = tableWidth / 5;
      MyUtils.drawString("segment",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("T",
                         diagramTableSpace + leftSpace + diagramWidth + colWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("V",
                         diagramTableSpace + leftSpace + diagramWidth + 2 * colWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("L = T  V",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("E = T + V",
                         diagramTableSpace + leftSpace + diagramWidth +
                         4 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);

      MyUtils.drawString("#",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth + colWidth +
                         colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         2 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2); //unicode for delta
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         4 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);

      for (i = 1; i <= numOfLayers; i++) {
        s = String.valueOf(i);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(kinEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(potEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 2 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(kinEnergy(i) - potEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 3 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff(kinEnergy(i) + potEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 4 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
      }

      MyUtils.drawString("\u0394t = ".concat(MyUtils.roundOff(deltaT, 2)).
                         concat(" sec"),
                         leftSpace + diagramWidth + 2 * colWidth,
                         topSpace + diagramHeight + 40, 2, f1, g2);

    }
    else if (appletMode == 4) { //accleration, T+V, action on segments and segment pairs and minima
      colWidth = tableWidth / 6;
      MyUtils.drawString("segment",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("a",
                         diagramTableSpace + leftSpace + diagramWidth + 1 * colWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("T + V",
                         diagramTableSpace + leftSpace + diagramWidth +
                         2 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("\u0394S",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("        k",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height, 2,
                         new Font("SansSerif", Font.PLAIN, 9), g2);

      MyUtils.drawString("\u0394S+\u0394S  ",
                         diagramTableSpace + leftSpace + diagramWidth +
                         4 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2); //unicode for delta
      MyUtils.drawString("        k      k+1",
                         diagramTableSpace + leftSpace + diagramWidth +
                         4 * colWidth + colWidth / 2, height, 2,
                         new Font("SansSerif", Font.PLAIN, 9), g2);

      MyUtils.drawString("min",
                         diagramTableSpace + leftSpace + diagramWidth +
                         5 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2); //unicode for delta

      MyUtils.drawString("k",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("meters/sec  ",
                         diagramTableSpace + leftSpace + diagramWidth +
                         1 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("                     2",
                         diagramTableSpace + leftSpace + diagramWidth +
                         1 * colWidth + colWidth / 2, height + height / 2 - 3,
                         2, new Font("SansSerif", Font.PLAIN, 9), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         2 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         3 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         4 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2); //unicode for delta
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         5 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);

      for (i = 1; i <= numOfLayers; i++) {
        s = MyUtils.roundOff(kinEnergy(i) + potEnergy(i), 1);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 2 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = String.valueOf(i);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff( (kinEnergy(i) - potEnergy(i)) * deltaT, 4);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 3 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        if (i != numOfLayers) {
          s = MyUtils.roundOff(acceleration(i), 1);
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth +
                             1 * colWidth + colWidth / 2,
                             (i + 2) * height + height / 2, 2, f1, g2);
          s = MyUtils.roundOff(sAPlusSB(i), 4);
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth +
                             4 * colWidth + colWidth / 2,
                             (i + 2) * height + height / 2, 2, f1, g2);
          s = MyUtils.roundOff(sABMin[i - 1], 4);
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth +
                             5 * colWidth + colWidth / 2,
                             (i + 2) * height + height / 2, 2, f1, g2);
        }
      }

      MyUtils.drawString("\u0394t = ".concat(MyUtils.roundOff(deltaT, 2)).
                         concat(" sec"),
                         leftSpace + diagramWidth + 2 * colWidth,
                         topSpace + diagramHeight + 40, 2, f1, g2);

    }
    else if (appletMode == 5) { //just actions on segments and their total and sum of dS on pair of succesive segments corresponding to the intermediate point just being changed
      colWidth = tableWidth / 4;
      height = 3 * height / 2;
      int numDig = 6; //number of digits for displaying delta S

      MyUtils.drawString("segment",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);
      MyUtils.drawString("\u0394S",
                         diagramTableSpace + leftSpace + diagramWidth +
                         1 * colWidth + colWidth / 2, height / 2, 2,
                         new Font("SansSerif", Font.BOLD, 12), g2);

      MyUtils.drawString("#",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("joules",
                         diagramTableSpace + leftSpace + diagramWidth +
                         1 * colWidth + colWidth / 2, height + height / 2, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);

      for (i = 1; i <= numOfLayers; i++) {
        s = String.valueOf(i);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
        s = MyUtils.roundOff( (kinEnergy(i) - potEnergy(i)) * deltaT, numDig);
        MyUtils.drawString(s,
                           diagramTableSpace + leftSpace + diagramWidth + 1 * colWidth +
                           colWidth / 2, (i + 2) * height, 2, f1, g2);
      }

      g2.drawLine(diagramTableSpace + leftSpace + diagramWidth + 1 * colWidth +
                  colWidth / 2 - 30, (numOfLayers + 3) * height - 5,
                  diagramTableSpace + leftSpace + diagramWidth + 1 * colWidth +
                  colWidth / 2 + 30, (numOfLayers + 3) * height - 5);
      MyUtils.drawString(MyUtils.roundOff(S, 4),
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth + colWidth / 2,
                         topSpace + (numOfLayers + 3) * height, 2, f1, g2);
      MyUtils.drawString("S Total",
                         diagramTableSpace + leftSpace + diagramWidth +
                         colWidth + colWidth / 2 + 35,
                         topSpace + (numOfLayers + 3) * height, 3, f1, g2);

      MyUtils.drawString("\u0394t = ".concat(MyUtils.roundOff(deltaT, 2)).
                         concat(" sec"),
                         diagramTableSpace + leftSpace + diagramWidth + 30,
                         topSpace + diagramHeight + 80, 3, f1, g2);

      //drawing rectangle with the dS(k)+dS(k+1) for k corresponding to intermediate point being changed
      if (drawTableRectangle) {
        if ( (kD != 0) && (kD != numOfLayers)) {
          g2.drawRect(diagramTableSpace + leftSpace + diagramWidth + 30,
                      (kD + 2) * height - height / 2 + 3, 4 * colWidth - 31,
                      2 * height);
          //headings
          String s1, s2;
          s1 = String.valueOf(kD);
          s2 = String.valueOf(kD + 1);

          MyUtils.drawString("\u0394S  +  \u0394S  ",
                             diagramTableSpace + leftSpace + diagramWidth +
                             2 * colWidth + colWidth / 2,
                             (1) * height - height / 2, 2,
                             new Font("SansSerif", Font.BOLD, 12), g2);
          MyUtils.drawString(s1.concat("          ").concat(s2),
                             diagramTableSpace + leftSpace + diagramWidth +
                             2 * colWidth + colWidth / 2 + 9,
                             (1) * height + 7 - height / 2, 2,
                             new Font("SansSerif", Font.BOLD, 9), g2);
          MyUtils.drawString("joules",
                             diagramTableSpace + leftSpace + diagramWidth +
                             2 * colWidth + colWidth / 2,
                             (1 + 1) * height - height / 2, 2,
                             new Font("SansSerif", Font.PLAIN, 12), g2);

          MyUtils.drawString("(\u0394S  +  \u0394S  )",
                             diagramTableSpace + leftSpace + diagramWidth +
                             3 * colWidth + colWidth / 2 - 7,
                             (1) * height - height / 2, 2,
                             new Font("SansSerif", Font.BOLD, 12), g2);
          MyUtils.drawString("joules",
                             diagramTableSpace + leftSpace + diagramWidth +
                             3 * colWidth + colWidth / 2,
                             (1 + 1) * height - height / 2, 2,
                             new Font("SansSerif", Font.PLAIN, 12), g2);
          MyUtils.drawString(s1.concat("          ").concat(s2),
                             diagramTableSpace + leftSpace + diagramWidth +
                             3 * colWidth + colWidth / 2 + 9 - 7,
                             (1) * height + 7 - height / 2, 2,
                             new Font("SansSerif", Font.BOLD, 9), g2);
          MyUtils.drawString("min",
                             diagramTableSpace + leftSpace + diagramWidth +
                             3 * colWidth + colWidth / 2 + 35,
                             (1) * height + 7 - height / 2, 2,
                             new Font("SansSerif", Font.BOLD, 9), g2);

          //values
          s = MyUtils.roundOff(sAPlusSB(kD), numDig);
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth +
                             2 * colWidth + colWidth / 2,
                             (kD + 2) * height + height / 2, 2,
                             new Font("SansSerif", Font.PLAIN, 12), g2);
          s = MyUtils.roundOff(sABMin[kD - 1], numDig);
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth +
                             3 * colWidth + colWidth / 2,
                             (kD + 2) * height + height / 2, 2,
                             new Font("SansSerif", Font.PLAIN, 12), g2);
          s = MyUtils.roundOff(acceleration(kD), 1);
          s = "Acceleration on segment pair ".concat(s1).concat("-").concat(s2).
              concat(":   ").concat(s).concat(" meters/sec");
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth + 30,
                             topSpace + diagramHeight, 3,
                             new Font("SansSerif", Font.BOLD, 12), g2);
          int sLength = g2.getFontMetrics().stringWidth(s);
          MyUtils.drawString("2",
                             diagramTableSpace + leftSpace + diagramWidth + sLength +
                             30 + 3, topSpace + diagramHeight - height / 2 + 7,
                             2, new Font("SansSerif", Font.BOLD, 9), g2);

          s = MyUtils.roundOff(kinEnergy(kD) + potEnergy(kD), 1);
          s = "Energy on segment ".concat(s1).concat(":   ").concat(s).concat(
              " joules");
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth + 30,
                             topSpace + diagramHeight + height, 3,
                             new Font("SansSerif", Font.BOLD, 12), g2);
          s = MyUtils.roundOff(kinEnergy(kD + 1) + potEnergy(kD + 1), 1);
          s = "Energy on segment ".concat(s2).concat(":   ").concat(s).concat(
              " joules");
          MyUtils.drawString(s,
                             diagramTableSpace + leftSpace + diagramWidth + 30,
                             topSpace + diagramHeight + 2 * height, 3,
                             new Font("SansSerif", Font.BOLD, 12), g2);
        }
      }

      //drawing segment numbers at top of diagram
      if (!isZoomed) {
        for (i = 1; i <= numOfLayers; i++) {
          s = String.valueOf(i);
          MyUtils.drawString(s, sc.xToi( (i - 0.5) * deltaT), topSpace + 10, 2,
                             new Font("SansSerif", Font.PLAIN, 9), g2);
        }
      }
      else { //if it is zoomed
        int k1 = Math.round( (float) Math.floor( (sc.iTox(leftSpace) - t1) /
                                                deltaT));
        int k2 = Math.round( (float) Math.floor( (sc.iTox(leftSpace +
            diagramWidth - 1) - t1) / deltaT));
        int ii;
        s = String.valueOf(k1 + 1);
        if (k1 == k2) {
          ii = sc.xToi( (sc.iTox(leftSpace) + sc.iTox(leftSpace + diagramWidth)) /
                       2.0);
          MyUtils.drawString(s, ii, topSpace + 10, 2,
                             new Font("SansSerif", Font.PLAIN, 9), g2);
        }
        else {
          ii = sc.xToi( (sc.iTox(leftSpace) + (k1 + 1) * deltaT + t1) / 2.0);
          MyUtils.drawString(s, ii, topSpace + 10, 2,
                             new Font("SansSerif", Font.PLAIN, 9), g2);
          ii = sc.xToi( (sc.iTox(leftSpace + diagramWidth) + (k2) * deltaT + t1) /
                       2.0);
          s = String.valueOf(k2 + 1);
          MyUtils.drawString(s, ii, topSpace + 10, 2,
                             new Font("SansSerif", Font.PLAIN, 9), g2);
        }
        for (int k = k1 + 2; k <= k2; k++) {
          s = String.valueOf(k);
          MyUtils.drawString(s, sc.xToi( (k - 0.5) * deltaT), topSpace + 10, 2,
                             new Font("SansSerif", Font.PLAIN, 9), g2);
        }
      }
    }

  }

  public void myDrawLine(double x1, double y1, double x2, double y2,
                         Graphics2D g2) {
    double X1 = sc.iTox(leftSpace);
    double X2 = sc.iTox(leftSpace + diagramWidth);
    double Y1 = sc.jToy(topSpace + diagramHeight);
    double Y2 = sc.jToy(topSpace);
    Point2D p1, p2, pA, pB;
    p1 = new Point2D(x1, y1);
    p2 = new Point2D(x2, y2);

    if (isInside(p1, X1, Y1, X2, Y2)) {
      if (isInside(p2, X1, Y1, X2, Y2)) {
        //p1 & p2 inside
        pA = p1;
        pB = p2;
        g2.drawLine(sc.xToi(pA.getX()), sc.yToj(pA.getY()), sc.xToi(pB.getX()),
                    sc.yToj(pB.getY()));
      }
      else {
        //p1 inside, p2 outside
        handleDrawingP1P2(p1, p2, X1, Y1, X2, Y2, g2);
      }
    }
    else {
      if (isInside(p2, X1, Y1, X2, Y2)) {
        //p1 outside, p2 inside
        handleDrawingP1P2(p2, p1, X1, Y1, X2, Y2, g2);
      }
      else {
        //p1 outside, p2 outside
        handleDrawingP1P2(p1, p2, X1, Y1, X2, Y2, g2);
      }
    }
  }

  public boolean isInside(Point2D p, double X1, double Y1, double X2, double Y2) {
    double tol = 1e-6;
    return (p.getX() < X2 + tol) && (p.getX() > X1 - tol) &&
        (p.getY() > Y1 - tol) && (p.getY() < Y2 + tol);
  }

  public void handleDrawingP1P2(Point2D p1, Point2D p2, double X1, double Y1,
                                double X2, double Y2, Graphics2D g2) {
    //Used when one or both points p1 and p2 lie OUTSIDE rectangle.
    //If there is intersection of p1p2 with the rectangle, it is drawn here
    Point2D[] p = new Point2D[4];
    double x1 = p1.getX();
    double x2 = p2.getX();
    double y1 = p1.getY();
    double y2 = p2.getY();
    double xA, yA, xB, yB;
    double tol = 1e-6;
    double t;
    int k = 0;

    Point2D p0, pA, pB;

    //intersection with line y=Y1
    if (Math.abs(y2 - y1) > tol) {
      t = (Y1 - y1) / (y2 - y1);
      if ( (0 < t) && (t < 1)) {
        p0 = new Point2D(x1 + (x2 - x1) * t, y1 + (y2 - y1) * t);
        if (isInside(p0, X1, Y1, X2, Y2)) {
          p[k] = p0;
          k++;
        }
      }
    }
    //intersection with line y=Y2
    if (Math.abs(y2 - y1) > tol) {
      t = (Y2 - y1) / (y2 - y1);
      if ( (0 < t) && (t < 1)) {
        p0 = new Point2D(x1 + (x2 - x1) * t, y1 + (y2 - y1) * t);
        if (isInside(p0, X1, Y1, X2, Y2)) {
          p[k] = p0;
          k++;
        }
      }
    }
    //intersection with line x=X1
    if (Math.abs(x2 - x1) > tol) {
      t = (X1 - x1) / (x2 - x1);
      if ( (0 < t) && (t < 1)) {
        p0 = new Point2D(x1 + (x2 - x1) * t, y1 + (y2 - y1) * t);
        if (isInside(p0, X1, Y1, X2, Y2)) {
          p[k] = p0;
          k++;
        }
      }
    }
    //intersection with line x=X2
    if (Math.abs(x2 - x1) > tol) {
      t = (X2 - x1) / (x2 - x1);
      if ( (0 < t) && (t < 1)) {
        p0 = new Point2D(x1 + (x2 - x1) * t, y1 + (y2 - y1) * t);
        if (isInside(p0, X1, Y1, X2, Y2)) {
          p[k] = p0;
          k++;
        }
      }
    }
    if (k == 1) {
      pA = p1;
      pB = p[0];
      g2.drawLine(sc.xToi(pA.getX()), sc.yToj(pA.getY()), sc.xToi(pB.getX()),
                  sc.yToj(pB.getY()));
    }

    if (k == 2) {
      pA = p[0];
      pB = p[1];
      g2.drawLine(sc.xToi(pA.getX()), sc.yToj(pA.getY()), sc.xToi(pB.getX()),
                  sc.yToj(pB.getY()));
    }

  }

  public boolean intPointInside(int ii, int jj) {
    return (ii >= (leftSpace)) && (ii <= (leftSpace + diagramWidth)) &&
        (jj >= (topSpace)) && (jj <= (topSpace + diagramHeight));
  }

  public int getNumDigits(int exp) {
    if (exp >= 0)
      return 1;
    else {
      int n = Math.round( (float) ( -exp * Math.log(5) / Math.log(10)));
      if (n > 2)
        return 2;
      else
        return n;
    }
  }

  public double velocity(int i) {
    return (h[i] - h[i - 1]) / deltaT;
  }

  public double deltaVel(int i) {
    double v1, v2;
    v1 = (h[i] - h[i - 1]) / deltaT;
    v2 = (h[i + 1] - h[i]) / deltaT;
    return (v2 - v1);
  }

  public double acceleration(int i) {
    double a, v1, v2;
    v1 = (h[i] - h[i - 1]) / deltaT;
    v2 = (h[i + 1] - h[i]) / deltaT;
    a = (v2 - v1) / deltaT;
    return a;
  }

  public double sAPlusSB(int i) {
    return (kinEnergy(i + 1) - potEnergy(i + 1) + kinEnergy(i) - potEnergy(i)) *
        deltaT;
  }

  public double energy(int i) {
    double en;
    en = 0.5 * m * Math.pow( (h[i] - h[i - 1]) / deltaT, 2) +
        0.5 * m * g * (h[i] + h[i - 1]);
    return en;
  }

  public double kinEnergy(int i) {
    double en;
    en = 0.5 * m * Math.pow( (h[i] - h[i - 1]) / deltaT, 2);
    return en;
  }

  public double potEnergy(int i) {
    double en;
    en = 0.5 * m * g * (h[i] + h[i - 1]);
    return en;
  }

  public double action() {
    double S = 0;
    for (int i = 0; i < numOfLayers; i++) {
      S += 0.5 * m *
          (Math.pow(h[i + 1] - h[i], 2) / deltaT -
           g * (h[i] + h[i + 1]) * deltaT);
    }
    return S;
  }

  public void makeOneIteration() {
    //Gauss-Seidel method is used here...
    //Iteration eqns are derived from Euler's Variational Method
    for (int i = 1; i < numOfLayers; i++) {
      h[i] = (h[i + 1] + h[i - 1] + g * deltaT * deltaT) / 2.0;
    }
  }

  public void run() {
    while (!isNotHunting) {
      makeOneIteration();
      if ( (appletMode == 4) || (appletMode == 5)) {
        // adjusting values for sABMin[] to NEW starting minimal values
        sABMin = new double[numOfLayers - 1];
        for (int i = 0; i < numOfLayers - 1; i++) {
          sABMin[i] = sAPlusSB(i + 1);
        }
      }

      canvas.repaint();

      try {
        thread.sleep(10);
      }
      catch (InterruptedException exception) {
        System.err.println("Thread interrupted...");
      }
    }
  }

  public void mousePressed(MouseEvent e) {
  }

  public void mouseDragged(MouseEvent e) {
    int i, j;
    i = e.getX();
    j = e.getY();
    int tol = 4;

    if (isOver(i, j, 50) && isNotHunting && !isZoomDrag &&
        (j <= topSpace + diagramHeight) && (j >= topSpace) &&
        (i >= leftSpace - tol) && (i <= (leftSpace + diagramWidth + tol))) {
      canvas.setCursor(crHand);
      int k = Math.round( (float) ( (sc.iTox(i) - t1) / deltaT)); //index of point above which we are
      h[k] = sc.jToy(j);
      drawTableRectangle = true;
      cursorOverDiagram = true;

      if ( (appletMode == 4) || (appletMode == 5)) {
        //adjusting deltaS minima
        if (k - 2 >= 0)
          sABMin[k - 2] = sAPlusSB(k - 1);
        if (k + 1 < numOfLayers)
          sABMin[k] = sAPlusSB(k + 1);
        if ( (k > 0) && (k < numOfLayers))
          if (sAPlusSB(k) < sABMin[k - 1])
            sABMin[k - 1] = sAPlusSB(k);
      }

      //real coordinates below cursor
      tC = t1 + k * deltaT;
      hC = h[k];
      if ( (k == 0) || (k == numOfLayers)) {
        rescale(h[0], h[numOfLayers]);
        SMin = action();
      }
      kD = Math.round( (float) ( (tC - t1) / deltaT)); //index of intermediate point being dragged

      canvas.repaint();
    }
    else if (!isZoomed && isZoomDrag && isOverZoomRect(i, j, 200)) {
      zoomCenterX = sc.iTox(i);
      zoomCenterY = sc.jToy(j);
      if (zoomCenterX < zoomRectW / 2)
        zoomCenterX = zoomRectW / 2;
      if (zoomCenterX > t2 - zoomRectW / 2)
        zoomCenterX = t2 - zoomRectW / 2;
      if (zoomCenterY < zoomRectH / 2)
        zoomCenterY = zoomRectH / 2;
      if (zoomCenterY > 1.2 * hMax - zoomRectH / 2)
        zoomCenterY = 1.2 * hMax - zoomRectH / 2;

      canvas.setCursor(crHand);
      cursorOverDiagram = true;
      tC = sc.iTox(i);
      hC = sc.jToy(j);
      canvas.repaint();
    }
    else {
      if ( (j > topSpace + diagramHeight + tol) || (j < topSpace - tol) ||
          (i < leftSpace - tol) || (i > (leftSpace + diagramWidth + tol))) {
        canvas.setCursor(crDefault);
        cursorOverDiagram = false;
        canvas.repaint();
      }
      else {
        canvas.setCursor(crCrossHair);
        cursorOverDiagram = true;
        tC = sc.iTox(i);
        hC = sc.jToy(j);
        canvas.repaint();
      }
    }
  }

  public void mouseMoved(MouseEvent e) {
    int i, j;
    int tol = 4;
    i = e.getX();
    j = e.getY();
    //drawTableRectangle=false;
    if ( (j > topSpace + diagramHeight + tol) || (j < topSpace - tol) ||
        (i < leftSpace - tol) || (i > (leftSpace + diagramWidth + tol))) {
      canvas.setCursor(crDefault);
      cursorOverDiagram = false;
    }
    else
    if (isOver(i, j, 4) && !isZoomDrag && isNotHunting) {
      canvas.setCursor(crHand);
      cursorOverDiagram = true;
      int k = Math.round( (float) ( (sc.iTox(i) - t1) / deltaT)); //index of point we are above
      //real coordinates below cursor
      tC = t1 + k * deltaT;
      hC = h[k];
    }
    else if (!isZoomed && isZoomDrag && isNotHunting && isOverZoomRect(i, j, 4)) {
      canvas.setCursor(crHand);
      cursorOverDiagram = true;
    }
    else {
      canvas.setCursor(crCrossHair);
      cursorOverDiagram = true;
      //real coordinates below cursor
      tC = sc.iTox(i);
      hC = sc.jToy(j);
    }
    canvas.repaint();
  }

  public void mouseReleased(MouseEvent e) {
  }

  public void mouseEntered(MouseEvent e) {
  }

  public void mouseExited(MouseEvent e) {
    cursorOverDiagram = false;
  }

  public void mouseClicked(MouseEvent e) {
    int i, j;
    int tol = 4;

    i = e.getX();
    j = e.getY();

    if (isOver(i, j, 10) && !isZoomDrag && (j <= topSpace + diagramHeight + tol) &&
        (j >= topSpace - tol) && (i >= leftSpace - tol) &&
        (i <= (leftSpace + diagramWidth + tol))) {
      kD = Math.round( (float) ( (sc.iTox(i) - t1) / deltaT)); //index of point above which we are
      drawTableRectangle = true;
      canvas.repaint();
    }
  }

  public boolean isOverZoomRect(int i, int j, int tol) {
    int iZC, jZC;
    iZC = sc.xToi(zoomCenterX);
    jZC = sc.yToj(zoomCenterY);
    return (Math.abs(i - iZC) < tol) && (Math.abs(j - jZC) < tol);
  }

  public boolean isOver(int i, int j, int tol) {
    boolean b = false;
    int i1, i2;

    if (movableEnds == 0) {
      i1 = 1;
      i2 = numOfLayers - 1;
    }
    else {
      i1 = 0;
      i2 = numOfLayers;
    }

    for (int k = i1; k <= i2; k++) {
      if ( (Math.abs(sc.xToi(t1 + k * deltaT) - i) < 4) &&
          (Math.abs(sc.yToj(h[k]) - j) < tol)) {
        b = true;
        break;
      }
    }
    return b;
  }

  public String getAppletInfo() {
    return
        "JAction - Java applet written by Slavomir Tuleja for Edwin F. Taylor";
  }
}

class Canvas
    extends JPanel {

  BufferedImage image;
  JAction applet;

  public Canvas(BufferedImage image, JAction applet) {
    this.image = image;
    this.applet = applet;
  }

  public void paintComponent(Graphics g) {
    Graphics2D g2 = (Graphics2D) g;
    g2.drawImage(image, null, 0, 0);
    applet.drawTopLayer(g2);
  }

}

class Point2D {
  double x, y;

  public Point2D(double x, double y) {
    this.x = x;
    this.y = y;
  }

  public double getX() {
    return x;
  }

  public double getY() {
    return y;
  }

  public String toString() {
    return new String(x + ", " + y);
  }
}

class ComboHandler
    implements ActionListener {
  private JAction applet;

  public ComboHandler(JAction applet) {
    this.applet = applet;
  }

  public void actionPerformed(ActionEvent e) {
    applet.updateZoom();
  }
}