package org.opensourcephysics.tuleja.energydisplay;

/**
 * <p>Title: Energy display</p>
 * <p>Description: One dimensional motion of a particle with energy E in a potential</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Slavomir Tuleja
 *
 * Java applet JEnergyDisplay simulates one dimensional motion of a particle
 * with known energy E in a potential.
 * Programmed by Slavomir Tuleja
 * Advised by Edwin F. Taylor
 * Project uses classes of Open Source Physics project www.opensourcephysics.org
 * Work started on December 29, 2003
 * -> version 0.01 sent to Edwin on December 31, 2003
 * -> Work on version 0.02 started on January 2, 2004
 * -> version 0.02 sent to Edwin on January 2, 2004
 * -> Work on version 0.03 started on January 9, 2004
 * -> version 0.03 sent to Edwin on January 11, 2004
 * -> Work on version 0.04 started on January 24, 2004
 * -> version 0.04 sent to Edwin on February 2, 2004
 * -> Work on version 0.05 started on February 7, 2004
 * -> version 0.05 sent to Edwin on February 15, 2004
 * -> Work on version 0.06 started on June 1, 2004
 * -> version 0.06 sent to Edwin on June 1, 2004
 * -> Work on version 0.07 started on June 2, 2004
 * -> version 0.07 sent to Edwin on June 2, 2004
 * -> Work on version 0.08 started on June 6, 2004
 * -> version 0.08 sent to Edwin on June 9, 2004
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;
import org.opensourcephysics.numerics.*;
import java.util.*;
import java.io.*;

public class JEnergyDisplay
    extends JApplet
    implements ActionListener, Runnable, MouseListener, MouseMotionListener {

  Scale scEnergy, scSpacetime, scPhaseDiagram;
  BufferedImage image;
  Graphics2D g2;
  Canvas canvas;
  Container cp;
  JPanel pnlSouth, pnlPotential, pnlButtons, pnlMessages;
  JButton btnStartStopResume, btnReset, btnStepForward, btnStepBack;
  JButton btnPEStep1, btnPEStep2, btnPEStep3, btnPEStep4, btnPECont1,
      btnPECont2;
  JButton btnTitlePageContinue;
  JLabel lblMessage;
  JEditorPane editorPane;
  ImageIcon iconStep1, iconStep2, iconStep3, iconStep4, iconCont1, iconCont2;

  JMenuBar menuBar;
  JMenu menuBackground, menuPotential, menuDisplay;
  JMenuItem menuItemStep, menuItemSmooth, menuItemFormula, menuItemTitlePage,
      menuItemIntroduction, menuItemHelp, menuItemWorldline,
      menuItemPhaseDiagram;

  Thread thread = null;
  Potential U;
  EnergyODE ode;
  ODEAdaptiveSolver odeSolver;
  Object history;
  ListIterator it;

  int leftBlankSpace = 40;
  int blankSpace = 40;
  int width = 700, height = 500, heightE = 190;
  boolean running = false;

  boolean editPotential = false, editDot = false, editClockArrow = false;
  boolean isHotPointDragged = false, isUntouched = true;
  boolean isRescaling=false;

  double lastSelectedE;
  double lastSelectedX = 0;
  double lastV;
  double x, v, t;
  double period = 4;
  int numberOfSegments = 0;
  int totalNumberOfSegments = 100;
  final int DISPLAY_MODE_WORLDLINES = 0;
  final int DISPLAY_MODE_PHASE_DIAGRAM = 1;
  int displayMode = DISPLAY_MODE_WORLDLINES; //0 -- worldlines; 1 -- phase diagram
  int lastDirectionOfIteration;//1...forward; (-1)...back

  ListIterator historyIterator;

  int clockHeadI, clockHeadJ;
  int draggedHotPointIndex = 0;
  int numberOfRESETClicks = 0;
  int numOfCycles=0;
  float[] dashed = {
      2.0f, 2.0f};
  float[] solid = {
      12.0f, 0.0f};

  BasicStroke bsDashed = new BasicStroke(1.0f, BasicStroke.CAP_ROUND,
                                         BasicStroke.JOIN_MITER, 1.0f, dashed,
                                         0.0f);
  BasicStroke bsSolid = new BasicStroke(1.0f, BasicStroke.CAP_ROUND,
                                        BasicStroke.JOIN_MITER, 1.0f, solid,
                                        0.0f);
  BasicStroke bsBold = new BasicStroke(2.0f);
  BasicStroke bsDoubleBold = new BasicStroke(4.0f);


  /** Returns an ImageIcon, or null if the path was invalid. */
  protected ImageIcon createImageIcon(String path) {
    java.net.URL imgURL = JEnergyDisplay.class.getResource(path);
    if (imgURL != null) {
      return new ImageIcon(imgURL);
    }
    else {
      System.err.println("Couldn't find file: " + path);
      return null;
    }
  }

  public static void main(String[] args) {
    JFrame fr = new JFrame("Worldlines");
    fr.setSize(706, 593);
    JEnergyDisplay applet = new JEnergyDisplay();
    applet.init();
    applet.start();
    fr.getContentPane().add(applet, BorderLayout.CENTER);
    fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    fr.setVisible(true);
  }


  /**
   * Initializing method of the applet. Creates GUI and prepares some other objects.
   */
  public void init() {
    //creation of GUI components
    //setting Java look and feel
    try {
      UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
    }
    catch (Exception e) {
      System.err.println("Couldn't use the cross-platform " + "look and feel: " +
                         e);
    }

    //buttons
    btnStartStopResume = new JButton("Start");
    btnStartStopResume.addActionListener(this);
    btnReset = new JButton("Reset");
    btnReset.addActionListener(this);
    btnStepForward = new JButton(">>");
    btnStepForward.addActionListener(this);
    btnStepForward.setEnabled(false);
    btnStepBack = new JButton("<<");
    btnStepBack.addActionListener(this);
    btnStepBack.setEnabled(false);

    String imagePath = "/org/opensourcephysics/tuleja/energydisplay/resources/";
    iconStep1 = createImageIcon(imagePath+"step1.gif");
    iconStep2 = createImageIcon(imagePath+"step2.gif");
    iconStep3 = createImageIcon(imagePath+"step3.gif");
    iconStep4 = createImageIcon(imagePath+"step4.gif");
    iconCont1 = createImageIcon(imagePath+"cont1.gif");
    iconCont1 = createImageIcon(imagePath+"cont2.gif");


    btnPEStep1 = new JButton(iconStep1);
    btnPEStep1.setActionCommand("btnPEStep1");
    btnPEStep1.setBackground(Color.white);
    btnPEStep1.addActionListener(this);
    btnPEStep2 = new JButton(iconStep2);
    btnPEStep2.setActionCommand("btnPEStep2");
    btnPEStep2.setBackground(Color.white);
    btnPEStep2.addActionListener(this);
    btnPEStep3 = new JButton(iconStep3);
    btnPEStep3.setActionCommand("btnPEStep3");
    btnPEStep3.setBackground(Color.white);
    btnPEStep3.addActionListener(this);
    btnPEStep4 = new JButton(iconStep4);
    btnPEStep4.setActionCommand("btnPEStep4");
    btnPEStep4.setBackground(Color.white);
    btnPEStep4.addActionListener(this);
    btnPECont1 = new JButton(iconCont1);
    btnPECont1.setActionCommand("btnPECont1");
    btnPECont1.setBackground(Color.white);
    btnPECont1.addActionListener(this);
    btnPECont2 = new JButton(iconCont2);
    btnPECont2.setActionCommand("btnPECont2");
    btnPECont2.setBackground(Color.white);
    btnPECont2.addActionListener(this);

    btnTitlePageContinue = new JButton("Continue");
    btnTitlePageContinue.addActionListener(this);

    lblMessage = new JLabel();
    lblMessage.setBackground(Color.white);

    editorPane = new JEditorPane();
    editorPane.setEditable(false);
    java.net.URL titleURL = JEnergyDisplay.class.getResource(
        imagePath+"Title.html");
    if (titleURL != null) {
      try {
        editorPane.setPage(titleURL);
      }
      catch (IOException e) {
        System.err.println("Attempted to read a bad URL: " + titleURL);
      }
    }
    else {
      System.err.println("Couldn't find file: Title.html");
    }

    //menu
    menuBar = new JMenuBar();
    menuBackground = new JMenu("Background");
    menuBackground.setMnemonic('b');
    menuPotential = new JMenu("Potential");
    menuPotential.setMnemonic('p');
    menuDisplay = new JMenu("Display");
    menuDisplay.setMnemonic('d');
    menuItemStep = new JMenuItem("Step");
    menuItemStep.addActionListener(this);
    menuItemStep.setMnemonic('t');
    menuItemSmooth = new JMenuItem("Smooth");
    menuItemSmooth.setMnemonic('m');
    menuItemSmooth.addActionListener(this);
    menuItemFormula = new JMenuItem("Formula");
    menuItemFormula.setMnemonic('f');
    menuItemFormula.addActionListener(this);
    menuItemFormula.setEnabled(false);
    menuItemTitlePage = new JMenuItem("Title page");
    menuItemTitlePage.setMnemonic('t');
    menuItemTitlePage.addActionListener(this);
    menuItemIntroduction = new JMenuItem("Introduction");
    menuItemIntroduction.setMnemonic('i');
    menuItemIntroduction.addActionListener(this);
    menuItemIntroduction.setEnabled(false);
    menuItemHelp = new JMenuItem("Help");
    menuItemHelp.setMnemonic('h');
    menuItemHelp.addActionListener(this);
    menuItemHelp.setEnabled(false);
    menuItemWorldline = new JMenuItem("Worldline");
    menuItemWorldline.setMnemonic('w');
    menuItemWorldline.addActionListener(this);
    menuItemPhaseDiagram = new JMenuItem("Phase diagram");
    menuItemPhaseDiagram.setMnemonic('p');
    menuItemPhaseDiagram.addActionListener(this);

    menuPotential.add(menuItemStep);
    menuPotential.add(menuItemSmooth);
    menuPotential.add(menuItemFormula);
    menuBackground.add(menuItemTitlePage);
    menuBackground.add(menuItemIntroduction);
    menuBackground.add(menuItemHelp);
    menuDisplay.add(menuItemWorldline);
    menuDisplay.add(menuItemPhaseDiagram);
    menuDisplay.setEnabled(false);
    menuBar.add(menuBackground);
    menuBar.add(menuPotential);
    menuBar.add(menuDisplay);
    setJMenuBar(menuBar);

    image = new BufferedImage(width - 1, height - 1, BufferedImage.TYPE_INT_RGB);
    g2 = image.createGraphics();
    g2.setRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING,
                                            RenderingHints.VALUE_ANTIALIAS_ON));

    canvas = new Canvas(image, this);
    canvas.addMouseListener(this);
    canvas.addMouseMotionListener(this);

    cp = getContentPane();
    cp.setBackground(Color.white);

    createTitlePageLayout();

  }

  /**
       * Used just before the simulation is started to set scale objects, the history
   * object and for initial drawing.
   */
  public void initialSetup() {
    ode.setInitialConditions(lastSelectedX, lastSelectedE);
    if (displayMode == DISPLAY_MODE_WORLDLINES) {
      scEnergy = new Scale( -2, 2, 0, 5, leftBlankSpace, width - blankSpace,
                           height - heightE, height - blankSpace);
      scSpacetime = new Scale( -2, 2, 0, 4, leftBlankSpace,
                              width - blankSpace,
                              5, height - heightE - blankSpace);
    }
    else {
      scEnergy = new Scale( -2, 2, 0, 5, leftBlankSpace, width - blankSpace,
                           height - heightE, height - blankSpace);
      scSpacetime = new Scale( -2, 2, 0, period, leftBlankSpace,
                              width - blankSpace,
                              200, height - heightE - blankSpace);
      scPhaseDiagram = new Scale( -2, 2, -Math.sqrt( (2.0 / 1.0) * (5)),
                                 Math.sqrt( (2.0 / 1.0) * (5)), leftBlankSpace,
                                 width - blankSpace,
                                 5, 200 - blankSpace);
    }

    odeSolver.setStepSize(period / totalNumberOfSegments);

    history = new LinkedList();
    x = lastSelectedX;
    t = 0;
    v = lastV * ode.getInitialDirection();

    addToHistory();
    canvas.repaint();
  }

  public void createSimulationLayout() {
    //components added to cp
    pnlSouth = new JPanel();
    pnlButtons = new JPanel();
    pnlMessages = new JPanel();

    pnlMessages.setLayout(new GridLayout(1, 1));
    pnlMessages.add(lblMessage);
    pnlMessages.setBackground(Color.white);

    pnlButtons.setLayout(new GridLayout(1, 4));
    pnlButtons.add(btnStartStopResume);
    pnlButtons.add(btnStepBack);
    pnlButtons.add(btnStepForward);
    pnlButtons.add(btnReset);
    pnlButtons.setBackground(Color.white);

    pnlSouth.setLayout(new GridLayout(2, 1));
    pnlSouth.add(pnlMessages);
    pnlSouth.add(pnlButtons);

    cp.setLayout(new BorderLayout());
    cp.add(canvas, BorderLayout.CENTER);
    cp.add(pnlSouth, BorderLayout.SOUTH);

    cp.validate();
    cp.repaint();
  }

  public void createStepPotentialLayout() {
    cp.setLayout(new BorderLayout());
    pnlPotential = new JPanel();
    pnlPotential.setLayout(new GridLayout(2, 2));
    pnlPotential.add(btnPEStep1);
    pnlPotential.add(btnPEStep2);
    pnlPotential.add(btnPEStep3);
    pnlPotential.add(btnPEStep4);
    lblMessage.setText("Click on one potential.");
    lblMessage.setHorizontalAlignment(JLabel.CENTER);
    cp.add(pnlPotential, BorderLayout.CENTER);
    cp.add(lblMessage, BorderLayout.SOUTH);

    cp.validate();
    cp.repaint();
  }

  public void createContPotentialLayout() {
    cp.setLayout(new BorderLayout());
    pnlPotential = new JPanel();
    pnlPotential.setLayout(new GridLayout(2, 1));
    pnlPotential.add(btnPECont1);
    pnlPotential.add(btnPECont2);
    lblMessage.setText("Click on one potential.");
    lblMessage.setHorizontalAlignment(JLabel.CENTER);
    cp.add(pnlPotential, BorderLayout.CENTER);
    cp.add(lblMessage, BorderLayout.SOUTH);

    cp.validate();
    cp.repaint();
  }

  public void createTitlePageLayout() {
    cp.setLayout(new BorderLayout());

    cp.add(editorPane, BorderLayout.CENTER);
    cp.add(btnTitlePageContinue, BorderLayout.SOUTH);

    cp.validate();
    cp.repaint();
  }

  /**
   * Computes and returns the period of motion.
   * @param x0 initial position
   * @param E energy
   * @return period T
   */
  private synchronized double getPeriod(double x0, double E) {
    double xOld = x0, tOld = 0, x = x0, time = 0;
    int numberOfPasses = 0;

    ode.setInitialConditions(x0, E);
    odeSolver.setStepSize(0.004);

    while (numberOfPasses < 3) {
      xOld = ode.getX();
      tOld = ode.getT();
      time += odeSolver.step();

      x = ode.getX();

      if ( (x0 - xOld) * (x - x0) >= 0) {
        numberOfPasses++;
      }
    }

    //linear interpolation...
    if (Math.abs(x - xOld) < 1e-7)
      return 1;
    else
      return time - (time - tOld) * (x - x0) / (x - xOld);
  }

  /**
   * Precomputes one cycle of periodic motion of the particle. The motion is
   * recorded in a linked list.
   */
  public synchronized void precomputeOneCycle() {
    double t = 0;
    double dt;
    period = getPeriod(lastSelectedX, lastSelectedE);
    initialSetup(); //initializes the history...
    ode.setInitialConditions(lastSelectedX, lastSelectedE);
    totalNumberOfSegments = 100;
    dt = period / totalNumberOfSegments;

    while (dt > 0.02) {
      dt = dt / 2.0;
      totalNumberOfSegments = totalNumberOfSegments * 2;
    }

    odeSolver.setStepSize(dt);
    odeSolver.setTolerance(1e-9);

    int k=0;
    while(k<totalNumberOfSegments){
      t += odeSolver.step();
      addToHistory();
      k++;
    }
  }

  /**
   * Defining what will happen after user clicks the buttons.
   * @param e action event
   */
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Start")) {
      isUntouched = false;
      btnStartStopResume.setText("Stop");
      btnReset.setEnabled(false);
      btnStepForward.setEnabled(false);
      btnStepBack.setEnabled(false);
      menuPotential.setEnabled(false);
      menuBackground.setEnabled(false);
      menuDisplay.setEnabled(false);

      if(displayMode==DISPLAY_MODE_PHASE_DIAGRAM) {
        isRescaling = true;
        draw();
        btnStartStopResume.setEnabled(false);
      }
      else {
        precomputeOneCycle();
        historyIterator = ( (java.util.List) history).listIterator();
      }

      running = true;
      thread = new Thread(this);
      thread.start();
    }
    else if (e.getActionCommand().equals("Stop")) {
      btnStartStopResume.setText("Resume");
      btnReset.setEnabled(true);
      btnStepForward.setEnabled(true);
      btnStepBack.setEnabled(true);
      running = false;
      lastDirectionOfIteration=1;
    }
    else if (e.getActionCommand().equals("Resume")) {
      btnStartStopResume.setText("Stop");
      btnReset.setEnabled(false);
      btnStepForward.setEnabled(false);
      btnStepBack.setEnabled(false);
      if(lastDirectionOfIteration==1){
        if (historyIterator.hasPrevious()){
          historyIterator.previous();
        }
        else{// !historyIterator.hasPrevious()
          historyIterator = ( (java.util.List) history).listIterator();
          while (historyIterator.hasNext()) {
            historyIterator.next();
          }
          historyIterator.previous();
        }
      }
      else{//lastDirectionOfIteration==-1
      }
      lastDirectionOfIteration=1;
      running = true;
      thread = new Thread(this);
      thread.start();
    }
    else if (e.getActionCommand().equals("Reset")) {
      btnStartStopResume.setText("Start");
      isUntouched = true;
      menuPotential.setEnabled(true);
      menuBackground.setEnabled(true);
      menuDisplay.setEnabled(true);
      btnStepForward.setEnabled(false);
      btnStepBack.setEnabled(false);
      numberOfRESETClicks++;
      initialSetup();
      numberOfSegments = 0;
      numOfCycles = 0;
    }
    else if (e.getActionCommand().equals("<<")) {
      SolutionDatum sd;
      if(displayMode==DISPLAY_MODE_PHASE_DIAGRAM){
        if(lastDirectionOfIteration==-1){
          if(historyIterator.hasPrevious()){
            //...
            sd = ( (SolutionDatum) historyIterator.previous());
            t = sd.getT();
            x = sd.getX();
            v = sd.getV();
            if(numberOfSegments<totalNumberOfSegments) numberOfSegments--;
            canvas.repaint();
          }
          else{// !historyIterator.hasPrevious()
            //...
            if(numberOfSegments==totalNumberOfSegments){
              historyIterator = ( (java.util.List) history).listIterator();
              while (historyIterator.hasNext()) {
                historyIterator.next();
              }
              historyIterator.previous();
              sd = ( (SolutionDatum) historyIterator.previous());
              t = sd.getT();
              x = sd.getX();
              v = sd.getV();
              canvas.repaint();
            }
            else{//numberOfSegments<totalNumberOfSegments
              //do nothing
            }
          }
        }
        else{//lastDirectionOfIteration==1
          historyIterator.previous();
          if(historyIterator.hasPrevious()){
            //...
            sd = ( (SolutionDatum) historyIterator.previous());
            t = sd.getT();
            x = sd.getX();
            v = sd.getV();
            if(numberOfSegments<totalNumberOfSegments) numberOfSegments--;
            canvas.repaint();
          }
          else{// !historyIterator.hasPrevious()
            //...
            if(numberOfSegments==totalNumberOfSegments){
              historyIterator = ( (java.util.List) history).listIterator();
              while (historyIterator.hasNext()) {
                historyIterator.next();
              }
              historyIterator.previous();
              sd = ( (SolutionDatum) historyIterator.previous());
              t = sd.getT();
              x = sd.getX();
              v = sd.getV();
              canvas.repaint();
            }
            else{//numberOfSegments<totalNumberOfSegments
              //do nothing
            }
          }

        }
      }
      else{// displayMode==DISPLAY_MODE_WORLDLINES
        if(lastDirectionOfIteration==-1){
          if(historyIterator.hasPrevious()){
            //...
            sd = ( (SolutionDatum) historyIterator.previous());
            t = sd.getT() + numOfCycles * period;
            x = sd.getX();
            v = sd.getV();
            numberOfSegments--;
            canvas.repaint();
          }
          else{// !historyIterator.hasPrevious()
            //...
            if(numOfCycles>0){
              historyIterator = ( (java.util.List) history).listIterator();
              while (historyIterator.hasNext()) {
                historyIterator.next();
              }
              historyIterator.previous();
              numOfCycles--;
              numberOfSegments--;
              sd = ( (SolutionDatum) historyIterator.previous());
              t = sd.getT() + numOfCycles * period;
              x = sd.getX();
              v = sd.getV();
              canvas.repaint();
            }
            else{//numOfCycles=0
              //do nothing
            }
          }
        }
        else{//lastDirectionOfIteration==1
          historyIterator.previous();
          if(historyIterator.hasPrevious()){
            //...
            sd = ( (SolutionDatum) historyIterator.previous());
            t = sd.getT() + numOfCycles * period;
            x = sd.getX();
            v = sd.getV();
            numberOfSegments--;
            canvas.repaint();
          }
          else{// !historyIterator.hasPrevious()
            //...
            if(numOfCycles>0){
              historyIterator = ( (java.util.List) history).listIterator();
              while (historyIterator.hasNext()) {
                historyIterator.next();
              }
              historyIterator.previous();
              numOfCycles--;
              numberOfSegments--;
              sd = ( (SolutionDatum) historyIterator.previous());
              t = sd.getT() + numOfCycles * period;
              x = sd.getX();
              v = sd.getV();
              canvas.repaint();
            }
            else{//numOfCycles=0
              //do nothing
            }
          }
        }
      }

      lastDirectionOfIteration=-1;
      isUntouched=false;

    }
    else if (e.getActionCommand().equals(">>")) {
      SolutionDatum sd;
      if(displayMode==DISPLAY_MODE_PHASE_DIAGRAM){
        if(lastDirectionOfIteration==1){
          if(historyIterator.hasNext()){
            //...
            sd = ( (SolutionDatum) historyIterator.next());
            t = sd.getT();
            x = sd.getX();
            v = sd.getV();
            if (numberOfSegments < totalNumberOfSegments)
              numberOfSegments++;
            canvas.repaint();
          }
          else{// !historyIterator.hasNext()
            //...
            historyIterator = ( (java.util.List) history).listIterator();
            historyIterator.next();
            sd = ( (SolutionDatum) historyIterator.next());
            t = sd.getT();
            x = sd.getX();
            v = sd.getV();
            if (numberOfSegments < totalNumberOfSegments)
              numberOfSegments++;
            canvas.repaint();
          }
        }
        else{//lastDirectionOfIteration==-1
          historyIterator.next();
          if(historyIterator.hasNext()){
            //..
            sd = ( (SolutionDatum) historyIterator.next());
            t = sd.getT();
            x = sd.getX();
            v = sd.getV();
            if (numberOfSegments < totalNumberOfSegments)
              numberOfSegments++;
            canvas.repaint();
          }
          else{// !historyIterator.hasNext()
            //..
            historyIterator = ( (java.util.List) history).listIterator();
            historyIterator.next();
            sd = ( (SolutionDatum) historyIterator.next());
            t = sd.getT();
            x = sd.getX();
            v = sd.getV();
            if (numberOfSegments < totalNumberOfSegments)
              numberOfSegments++;
            canvas.repaint();
          }
        }
      }
      else{//displayMode==DISPLAY_MODE_WORLDLINE
        if(lastDirectionOfIteration==1){
          if(historyIterator.hasNext()){
            //...
            sd = ( (SolutionDatum) historyIterator.next());
            t = sd.getT() + numOfCycles * period;
            x = sd.getX();
            v = sd.getV();
            numberOfSegments++;
            canvas.repaint();
          }
          else{// !historyIterator.hasNext()
            //...
            historyIterator = ( (java.util.List) history).listIterator();
            historyIterator.next();
            sd = ( (SolutionDatum) historyIterator.next());
            numOfCycles++;
            t = sd.getT() + numOfCycles * period;
            x = sd.getX();
            v = sd.getV();
            numberOfSegments++;
            canvas.repaint();
          }
        }
        else{//lastDirectionOfIteration==-1
          historyIterator.next();
          if(historyIterator.hasNext()){
            //..
            sd = ( (SolutionDatum) historyIterator.next());
            t = sd.getT() + numOfCycles * period;
            x = sd.getX();
            v = sd.getV();
            numberOfSegments++;
            canvas.repaint();
          }
          else{// !historyIterator.hasNext()
            //..
            historyIterator = ( (java.util.List) history).listIterator();
            historyIterator.next();
            sd = ( (SolutionDatum) historyIterator.next());
            numOfCycles++;
            t = sd.getT() + numOfCycles * period;
            x = sd.getX();
            v = sd.getV();
            numberOfSegments++;
            canvas.repaint();
          }
        }
      }

      lastDirectionOfIteration=1;
      isUntouched=false;
    }
    else if (e.getActionCommand().equals("Step")) {
      cp.removeAll();
      isUntouched = true;
      numberOfRESETClicks = 0;
      createStepPotentialLayout();
      menuDisplay.setEnabled(false);
    }
    else if (e.getActionCommand().equals("Smooth")) {
      cp.removeAll();
      isUntouched = true;
      numberOfRESETClicks = 0;
      createContPotentialLayout();
      menuDisplay.setEnabled(false);
    }
    else if (e.getActionCommand().equals("Formula")) {
      isUntouched = true;
      numberOfRESETClicks = 0;
    }
    else if (e.getActionCommand().equals("Title page")) {
      cp.removeAll();
      isUntouched = true;
      numberOfRESETClicks = 0;
      createTitlePageLayout();
      menuDisplay.setEnabled(false);
    }
    else if (e.getActionCommand().equals("Introduction")) {
      isUntouched = true;
      numberOfRESETClicks = 0;
      menuDisplay.setEnabled(false);
    }
    else if (e.getActionCommand().equals("Help")) {
      isUntouched = true;
      numberOfRESETClicks = 0;
      menuDisplay.setEnabled(false);
    }
    else if (e.getActionCommand().equals("Worldline")) {
      displayMode = DISPLAY_MODE_WORLDLINES;
      initialSetup();
      repaint();
    }
    else if (e.getActionCommand().equals("Phase diagram")) {
      displayMode = DISPLAY_MODE_PHASE_DIAGRAM;
      initialSetup();
      repaint();
    }
    else if (e.getActionCommand().equals("btnPEStep1")) {
      cp.removeAll();
      createSimulationLayout();
      menuDisplay.setEnabled(true);

      U = new PEStep(new double[] { -1.8, 0, 1.8}
                     , new double[] {0, 3, 3}
                     , new int[] {1, 0});
      ode = new EnergyODE(U);
      odeSolver = new RK45MultiStep(ode);
      odeSolver.setStepSize(0.02);
      lastSelectedE = U.evaluate(lastSelectedX) + 0.5;
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      initialSetup();
    }
    else if (e.getActionCommand().equals("btnPEStep2")) {
      cp.removeAll();
      createSimulationLayout();
      menuDisplay.setEnabled(true);

      U = new PEStep(new double[] { -1.8, -1, 1, 1.8}
                     , new double[] {0, 3, 3, 0}
                     , new int[] {1, 0, 2});
      ode = new EnergyODE(U);
      odeSolver = new RK45MultiStep(ode);
      odeSolver.setStepSize(0.02);
      lastSelectedE = U.evaluate(lastSelectedX) + 0.5;
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      initialSetup();
    }
    else if (e.getActionCommand().equals("btnPEStep3")) {
      cp.removeAll();
      createSimulationLayout();
      menuDisplay.setEnabled(true);

      U = new PEStep(new double[] { -1.8, -1, 0, 1, 1.8}
                     , new double[] {0, 3, 2, 3, 0}
                     , new int[] {1, 0, 0, 2});
      ode = new EnergyODE(U);
      odeSolver = new RK45MultiStep(ode);
      odeSolver.setStepSize(0.02);
      lastSelectedE = U.evaluate(lastSelectedX) + 0.5;
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      initialSetup();
    }
    else if (e.getActionCommand().equals("btnPEStep4")) {
      cp.removeAll();
      createSimulationLayout();
      menuDisplay.setEnabled(true);

      U = new PEStep(new double[] { -1.8, -1.5, -0.5, 0, 0.5, 1.5, 1.8}
                     , new double[] {0, 3, 3, 0, 3, 3, 0}
                     , new int[] {1, 0, 2, 1, 0, 2});
      ode = new EnergyODE(U);
      odeSolver = new RK45MultiStep(ode);
      odeSolver.setStepSize(0.02);
      lastSelectedE = U.evaluate(lastSelectedX) + 0.5;
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      initialSetup();
    }
    else if (e.getActionCommand().equals("btnPECont1")) {
      cp.removeAll();
      createSimulationLayout();
      menuDisplay.setEnabled(true);

      U = new PEContinuous(new double[] { -1, 0, 1}
                           , new double[] {0, 3.5, 0});
      ode = new EnergyODE(U);
      odeSolver = new RK45MultiStep(ode);
      odeSolver.setStepSize(0.02);
      lastSelectedE = U.evaluate(lastSelectedX) + 0.5;
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      initialSetup();
    }
    else if (e.getActionCommand().equals("btnPECont2")) {
      cp.removeAll();
      createSimulationLayout();
      menuDisplay.setEnabled(true);

      U = new PEContinuous(new double[] { -1.8, -1.5, -1.3, 0, 0.2, 1.8}
                           , new double[] {0, 0, 3, 3, 4, 4});
      ode = new EnergyODE(U);
      odeSolver = new RK45MultiStep(ode);
      odeSolver.setStepSize(0.02);
      lastSelectedE = U.evaluate(lastSelectedX) + 0.5;
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      initialSetup();
    }
    else if (e.getActionCommand().equals("Continue")) {
      cp.removeAll();
      createStepPotentialLayout();
    }
  }

  /**
   * This applet is Thread. Here we define what the thread should do if it is
   * in the run state.
   */
  public void run() {
    // This makes the applet wait for 2 seconds before it rescales the vertical axes
    // to make the user notice this rescaling -- WHEN displayMode==DISPLAY_MODE_PHASE_DIAGRAM
    if (isRescaling) {

      try {
        thread.sleep(2000);
      }
      catch (InterruptedException exception) {
        System.err.println("Thread interrupted...");
      }
      isRescaling = false;
      precomputeOneCycle();
      historyIterator = ( (java.util.List) history).listIterator();
      btnStartStopResume.setEnabled(true);
    }


    if(displayMode==DISPLAY_MODE_PHASE_DIAGRAM){
      SolutionDatum sd1, sd2;
      if (historyIterator.hasNext()) {
          sd1 = ( (SolutionDatum) historyIterator.next());

          while (running) {
            if(historyIterator.hasNext()) {
              sd2 = ( (SolutionDatum) historyIterator.next());
              t = sd2.getT();
              x = sd2.getX();
              v = sd2.getV();
              if (numberOfSegments < totalNumberOfSegments)
                numberOfSegments++;


              canvas.repaint();

              try {
                thread.sleep(50);
              }
              catch (InterruptedException exception) {
                System.err.println("Thread interrupted...");
              }

            }
            else{//!historyIterator.hasNext()
              historyIterator = ( (java.util.List) history).listIterator();
              sd1 = ( (SolutionDatum) historyIterator.next());
              continue;
            }
          }
      }
    }
    else{// displayMode==DISPLAY_MODE_WORLDLINE

      SolutionDatum sd1, sd2;
      if (historyIterator.hasNext()) {
          sd1 = ( (SolutionDatum) historyIterator.next());

          while (running) {
            if(historyIterator.hasNext()) {
              sd2 = ( (SolutionDatum) historyIterator.next());
              t = sd2.getT() + numOfCycles * period;
              x = sd2.getX();
              v = sd2.getV();
              numberOfSegments++;
              canvas.repaint();

              try {
                thread.sleep(50);
              }
              catch (InterruptedException exception) {
                System.err.println("Thread interrupted...");
              }

            }
            else{//!historyIterator.hasNext()
              historyIterator = ( (java.util.List) history).listIterator();
              numOfCycles++;
              sd1 = ( (SolutionDatum) historyIterator.next());
              continue;
            }
          }
        }
      }
  }

  /**
   * Adds a new solution point to the history. This method is
   * synchronized with the drawWorldline method since the two methods access the
   * same resources in the history LinkedList.
   */
  public synchronized void addToHistory() {
    ( (java.util.List) history).add(new SolutionDatum(ode.getX(), ode.getXDot(),
        ode.getT()));
  }

  /**
   * Draws the worldline. This method is
   * synchronized with the addToHistory method since the two methods access the
   * same resources in the history LinkedList.
   * @param g2 graphical context
   */
  public synchronized void drawWorldline(Graphics2D g2) {
    int k, numOfCycles=0;
    SolutionDatum sd1, sd2;
    int i1, i2, j1, j2;

    if(displayMode==DISPLAY_MODE_PHASE_DIAGRAM){
      k = 0;
      it = ( (java.util.List) history).listIterator();
      g2.setStroke(bsBold);
      if (it.hasNext()) {
        sd1 = ( (SolutionDatum) it.next());
        i1 = scSpacetime.xToi(sd1.getX());
        j1 = scSpacetime.yToj(sd1.getT());
        while (it.hasNext() && k < numberOfSegments) {
          sd2 = ( (SolutionDatum) it.next());
          i2 = scSpacetime.xToi(sd2.getX());
          j2 = scSpacetime.yToj(sd2.getT());
          k++;
          g2.drawLine(i1, j1, i2, j2);
          sd1 = sd2;
          i1 = i2;
          j1 = j2;
        }
      }
      g2.setStroke(bsSolid);
    }
    else{//displayMode==DISPLAY_MODE_WORLDLINE
      k = 0;
      it = ( (java.util.List) history).listIterator();
      g2.setStroke(bsBold);
      if (it.hasNext()) {
        sd1 = ( (SolutionDatum) it.next());
        i1 = scSpacetime.xToi(sd1.getX());
        j1 = scSpacetime.yToj(sd1.getT()+numOfCycles*period);
        while (k < numberOfSegments) {
          if(it.hasNext()){
            sd2 = ( (SolutionDatum) it.next());
            i2 = scSpacetime.xToi(sd2.getX());
            j2 = scSpacetime.yToj(sd2.getT()+numOfCycles*period);
            k++;
            g2.drawLine(i1, j1, i2, j2);
            sd1 = sd2;
            i1 = i2;
            j1 = j2;
          }
          else{
            numOfCycles++;
            it = ( (java.util.List) history).listIterator();
            sd1 = ( (SolutionDatum) it.next());
            i1 = scSpacetime.xToi(sd1.getX());
            j1 = scSpacetime.yToj(sd1.getT()+numOfCycles*period);
            continue;
          }
        }
      }
      g2.setStroke(bsSolid);

    }
  }

  /**
   * Draws the phase curve. This method is
   * synchronized with the addToHistory method since the two methods access the
   * same resources in the history LinkedList.
   * @param g2 graphical context
   */
  public synchronized void drawPhaseCurve(Graphics2D g2) {
    int k = 0;
    it = ( (java.util.List) history).listIterator();
    SolutionDatum sd1, sd2;
    g2.setStroke(bsBold);
    int i1, i2, j1, j2;
    if (it.hasNext()) {
      sd1 = ( (SolutionDatum) it.next());
      i1 = scPhaseDiagram.xToi(sd1.getX());
      j1 = scPhaseDiagram.yToj(sd1.getV());
      while (it.hasNext() && k < numberOfSegments) {
        sd2 = ( (SolutionDatum) it.next());
        i2 = scPhaseDiagram.xToi(sd2.getX());
        j2 = scPhaseDiagram.yToj(sd2.getV());
        k++;
        g2.drawLine(i1, j1, i2, j2);
        sd1 = sd2;
        i1 = i2;
        j1 = j2;
      }
    }
    g2.setStroke(bsSolid);
  }

  /**
   * Draws window graphical contents.
   */
  public synchronized void draw() {
    //white background
    g2.setColor(Color.white);
    g2.fillRect(0, 0, width - 1, height - 1);
    g2.setColor(Color.black);

    //energy axes
    drawAxes(g2, scEnergy, "Distance", "Energy", -2, 2, 0, 5, x, lastSelectedE, false, false);

    //spacetime axes
    if(displayMode == DISPLAY_MODE_PHASE_DIAGRAM){
      drawAxes(g2, scSpacetime, "Distance", "Time", -2, 2, 0, period, x,
               t, false, true);
    }
    else{//displayMode == DISPLAY_MODE_WORLDLINE
      drawAxes(g2, scSpacetime, "Distance", "Time", -2, 2, 0, 4, x,
               t, false, true);
    }

    //potential - each type of potential knows how to draw itself...
    U.draw(g2, scEnergy, btnStartStopResume.getText().equals("Start"),
           editPotential);

    //energy line
    g2.setStroke(bsBold);
    g2.setColor(new Color(0, 150, 0));
    g2.drawLine(scEnergy.xToi( -2), scEnergy.yToj(lastSelectedE),
                scEnergy.xToi(2),
                scEnergy.yToj(lastSelectedE));
    g2.setColor(Color.black);
    g2.setStroke(bsSolid);

    //gray line connecting energy dot and the worldline dot
    g2.setStroke(bsDashed);
    g2.setColor(Color.gray);
    g2.drawLine(scEnergy.xToi(x), scEnergy.yToj(lastSelectedE),
                scSpacetime.xToi(x), scSpacetime.yToj(t));
    g2.setColor(Color.black);
    g2.setStroke(bsSolid);

    //gray line connecting phase doagram dot and the worldline dot
    if (displayMode == DISPLAY_MODE_PHASE_DIAGRAM) {
      g2.setStroke(bsDashed);
      g2.setColor(Color.gray);
      g2.drawLine(scPhaseDiagram.xToi(x), scPhaseDiagram.yToj(v),
                  scSpacetime.xToi(x), scSpacetime.yToj(t));
      g2.setColor(Color.black);
      g2.setStroke(bsSolid);
    }

    //energy dot
    //this is done just for case that the energy dot is BELOW the potential when
    //potential is edited
    if (U.evaluate(lastSelectedX) > lastSelectedE) {
      lastSelectedE = U.evaluate(lastSelectedX);
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      initialSetup();
    }
    if (btnStartStopResume.getText().equals("Start")) {
      g2.setColor(Color.red);
      g2.fillOval(scEnergy.xToi(x) - 8, scEnergy.yToj(lastSelectedE) - 8,
                  16, 16);
      g2.setColor(Color.black);
      g2.drawOval(scEnergy.xToi(x) - 8, scEnergy.yToj(lastSelectedE) - 8,
                  16, 16);
    }
    else
      g2.fillOval(scEnergy.xToi(x) - 6, scEnergy.yToj(lastSelectedE) - 6,
                  12, 12);

    //direction arrow
    int dir;
    if (v > 0)
      dir = 1;
    else if (v < 0)
      dir = -1;
    else
      dir = 0;
    int magnitude = Math.round( (float) (blankSpace * v /
                                         ode.getXDotMax()));
    g2.setStroke(bsBold);
    g2.drawLine(scEnergy.xToi(x), scEnergy.yToj(lastSelectedE),
                scEnergy.xToi(x) + magnitude, scEnergy.yToj(lastSelectedE));

    GeneralPath arrowHead = new GeneralPath();
    arrowHead.moveTo(scEnergy.xToi(x) + magnitude, scEnergy.yToj(lastSelectedE));
    arrowHead.lineTo(scEnergy.xToi(x) + magnitude - dir * 12,
                     scEnergy.yToj(lastSelectedE) + 4);
    arrowHead.lineTo(scEnergy.xToi(x) + magnitude - dir * 12,
                     scEnergy.yToj(lastSelectedE) - 4);
    arrowHead.closePath();
    g2.fill(arrowHead);
    g2.draw(arrowHead);
    g2.setStroke(bsSolid);

    //worldline dot
    g2.setStroke(bsBold);
    g2.fillOval(scSpacetime.xToi(x) - 6,
                scSpacetime.yToj(t) - 6, 12, 12);

    //worldline
    drawWorldline(g2);

    //clock hand
    drawClock(g2);

    //phase diagram
    if (displayMode == DISPLAY_MODE_PHASE_DIAGRAM) {
      //phase diagram axes
      drawAxes(g2, scPhaseDiagram, "Distance", "Velocity", -2, 2,
               -Math.sqrt( (2.0 / 1.0) * (5)), Math.sqrt( (2.0 / 1.0) * (5)), x,
               v, false, true);

      //- 0 + near the vertical velocity axis
      Font f1 = new Font("SansSerif", Font.PLAIN, 12);
      g2.setFont(f1);
      int fheight = g2.getFontMetrics().getHeight(); //font height
      g2.setColor(Color.black);
      MyUtils.drawString("0", leftBlankSpace - 5, scPhaseDiagram.yToj(0) - 3, 2,
                         f1, g2);
      MyUtils.drawString("+", leftBlankSpace - 5,
                         scPhaseDiagram.yToj(Math.sqrt( (2.0 / 1.0) * (5))) - 3,
                         2, f1, g2);
      MyUtils.drawString("", leftBlankSpace - 5,
                         scPhaseDiagram.yToj( -Math.sqrt( (2.0 / 1.0) * (5))) -
                         3, 2, f1, g2);

      //phase diagram dot
      g2.setStroke(bsBold);
      g2.fillOval(scPhaseDiagram.xToi(x) - 6,
                  scPhaseDiagram.yToj(v) - 6, 12, 12);

      //phase curve
      drawPhaseCurve(g2);

    }

    //initial message
    if ( (isUntouched) && (numberOfRESETClicks < 2))
      if (numberOfRESETClicks == 1) {
        lblMessage.setText(
            "You may choose another potential from the potential menu.");
      }
      else {
        lblMessage.setText("<html>Click on any <font color=#ff0000>RED</font> object, then drag it in various directions.</html>");
      }
    else
      lblMessage.setText("");

    //rescaling message
    if(isRescaling){
      lblMessage.setText(
            "<html><font color=#0000ff>Time axis rescaled to span one cycle.</font></html>");
    }
    else
      lblMessage.setText("");


      notifyAll();
  }

  /**
   * Draws a half-clock with the initial slope arrow
   * @param g2 graphics context
   */
  public void drawClock(Graphics2D g2) {
    double x = lastSelectedX;
    double vMax = Math.sqrt(2 * (5 - U.evaluate(x)));
    double E = lastSelectedE;
    double v = ode.getInitialDirection() * Math.sqrt(2 * (E - U.evaluate(x)));
    double radius = 0.2;
    double lengthArrow = 0.35;
    //phi and phiMax are angles on screen
    double phiMax = Math.atan( (float) (scSpacetime.yToj(0) -
                                        scSpacetime.yToj(1)) /
                              (float) (scSpacetime.xToi(vMax) -
                                       scSpacetime.xToi(0)));
    double phi = Math.atan( (float) (scSpacetime.yToj(0) - scSpacetime.yToj(1)) /
                           (float)
                           Math.abs(scSpacetime.xToi(v) - scSpacetime.xToi(0)));
    if (ode.getInitialDirection() == -1) {
      phi = Math.PI - phi;
    }
    double wa = scSpacetime.xToi(2 * radius) - scSpacetime.xToi(0);
    double ha = wa;
    double xa = scSpacetime.xToi(x) - wa / 2.0;
    double ya = scSpacetime.yToj(0) - ha / 2.0;

    Arc2D pie = new Arc2D.Double(xa, ya, wa, wa, Math.toDegrees(phiMax),
                                 180.0 - 2 * Math.toDegrees(phiMax), Arc2D.PIE);
    g2.setColor(new Color(0.5f, 0.5f, 0.5f, 0.5f));
    g2.fill(pie);
    g2.setColor(Color.black);

    float length = (float) scSpacetime.xToi(x + lengthArrow) -
        (float) scSpacetime.xToi(x) - 5f; //length in pixels
    clockHeadI = scSpacetime.xToi(x) +
        Math.round( (float) (length * Math.cos(phi))); //global vaiables representing the pixel position of arrow head
    clockHeadJ = scSpacetime.yToj(0) -
        Math.round( (float) (length * Math.sin(phi)));

    //drawing the arrow with the red head
    AffineTransform at = new AffineTransform();
    at.rotate( -phi, scSpacetime.xToi(x), scSpacetime.yToj(0));
    g2.setTransform(at);
    g2.setStroke(bsBold);
    g2.drawLine(scSpacetime.xToi(x), scSpacetime.yToj(0),
                scSpacetime.xToi(x + lengthArrow), scSpacetime.yToj(0));

    GeneralPath arrowHead = new GeneralPath();
    arrowHead.moveTo( (float) scSpacetime.xToi(x + lengthArrow),
                     (float) scSpacetime.yToj(0));
    arrowHead.lineTo(scSpacetime.xToi(x + lengthArrow) - 20,
                     scSpacetime.yToj(0) - 8);
    arrowHead.lineTo(scSpacetime.xToi(x + lengthArrow) - 20,
                     scSpacetime.yToj(0) + 8);
    arrowHead.closePath();
    if (!btnStartStopResume.getText().equals("Start")) {
      g2.setColor(Color.BLACK);
    }
    else
      g2.setColor(Color.red);
    g2.fill(arrowHead);
    g2.setColor(Color.black);
    g2.draw(arrowHead);
    at.rotate( +phi, scSpacetime.xToi(x), scSpacetime.yToj(0));
    g2.setStroke(bsSolid);
    g2.setTransform(at);

    //drawing the "Initial Slope" label
    if ( (isUntouched) && (numberOfRESETClicks < 2)) {
      g2.setColor(Color.black);
      MyUtils.drawString("Initial", clockHeadI, clockHeadJ - 35, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
      MyUtils.drawString("slope", clockHeadI, clockHeadJ - 20, 2,
                         new Font("SansSerif", Font.PLAIN, 12), g2);
    }
  }

  /**
   * Draws axes using the specified g2.
   * @param g2 graphics context
   * @param sc scale object used in drawing
   * @param captionX caption for x-axis
   * @param captionY caption for y-axis
   * @param x1 x-value specifying the left beginning of the x-axis
   * @param x2 x-value specifying the right ending of the x-axis
   * @param y1 y-value specifying the bottom beginning of the y-axis
   * @param y2 y-value specifying the top ending of the y-axis
   * @param x current value of x-coordinate of a point drawn to the axes area. Used for automatic rescaling.
   * @param y current value of y-coordinate of a point drawn to the axes area. Used for automatic rescaling.
   * @param autoscaleX true if automatic rescaling of x-axis is allowed, false otherwise.
   * @param autoscaleY true if automatic rescaling of y-axis is allowed, false otherwise.
   */
  public void drawAxes(Graphics2D g2, Scale sc, String captionX,
                       String captionY, double x1, double x2, double y1,
                       double y2, double x, double y, boolean autoscaleX,
                       boolean autoscaleY) {
    double xMin = x1, xMax = x2;
    double yMin = y1, yMax = y2;
    double difference;
    if ( (autoscaleX) && (x >= xMax)) {
      sc.setYMax(x);
      xMax = x;
    }

    if ( (autoscaleY) && (y >= yMax)) {
      sc.setYMax(y);
      yMax = y;
    }

    //axis X ticks and numbers ...
    Font f = new Font("SansSerif", Font.PLAIN, 9);
    g2.setFont(f);
    double scaleDivisionX = 0.5 *
        Math.pow(2,
                 Math.round( (float) (Math.log(0.5 * (xMax - xMin)) /
                                      Math.log(2))));
    int digitsX = -Math.round( (float) (Math.log(0.02 * scaleDivisionX) /
                                        Math.log(10)));
    double xx = scaleDivisionX * Math.floor(xMin / scaleDivisionX);

    g2.setStroke(bsDashed);

    while (xx <= xMax) {
      g2.drawLine(sc.xToi(xx), sc.yToj(yMin), sc.xToi(xx), sc.yToj(yMax));
      //String s = MyUtils.roundOff(xx, digitsX);
      //MyUtils.drawString(s, sc.xToi(xx), sc.yToj(yMin) + 10, 2, f, g2);
      xx += scaleDivisionX;
    }

    //axisY ticks and numbers ...
    double scaleDivisionY = 0.5 *
        Math.pow(2,
                 Math.round( (float) (Math.log(0.5 * (yMax - yMin)) /
                                      Math.log(2))));
    int digitsY = -Math.round( (float) (Math.log(0.02 * scaleDivisionY) /
                                        Math.log(10)));
    double yy;
    if (yMin >= 0)
      yy = scaleDivisionY * Math.floor(yMin / scaleDivisionY);
    else
      yy = scaleDivisionY * (Math.floor(yMin / scaleDivisionY) + 1);

    while (yy <= yMax) {
      g2.drawLine(leftBlankSpace, sc.yToj(yy), width - blankSpace, sc.yToj(yy));
      //String s = MyUtils.roundOff(yy, digitsY);
      //MyUtils.drawString(s, leftBlankSpace - 5, sc.yToj(yy), 1, f, g2);
      yy += scaleDivisionY;
    }

    g2.setStroke(bsSolid);

    // axis X
    g2.setColor(Color.black);
    int i1 = sc.xToi(xMin);
    int i2 = sc.xToi(xMax);
    int j = sc.yToj(y1);

    //axis label X
    Font f1 = new Font("SansSerif", Font.PLAIN, 12);
    g2.setFont(f1);
    int fheight = g2.getFontMetrics().getHeight(); //font height
    g2.setColor(Color.black);
    MyUtils.drawString(captionX, (i1 + i2) / 2, j + fheight - 5, 2, f1, g2);

    //axis Y
    g2.setColor(Color.black);
    int j1 = sc.yToj(yMin);
    int j2 = sc.yToj(yMax);
    int i = leftBlankSpace;

    //axis label Y
    g2.setColor(Color.black);
    AffineTransform at = new AffineTransform();
    at.rotate( -Math.PI / 2.0, leftBlankSpace, (j1 + j2) / 2);
    g2.setTransform(at);
    MyUtils.drawString(captionY, leftBlankSpace, (j1 + j2) / 2 - fheight - 5,
                       2, f1, g2);
    at.rotate(Math.PI / 2.0, leftBlankSpace, (j1 + j2) / 2);
    g2.setTransform(at);

    //frame border
    g2.drawLine(i1, j1, i2, j1);
    g2.drawLine(i1, j2, i2, j2);
    g2.drawLine(i1, j1, i1, j2);
    g2.drawLine(i2, j1, i2, j2);

  }

  /**
   * Method of interface MouseListener.
   * @param e MouseEvent
   */
  public void mousePressed(MouseEvent e) {

  }

  /**
   * Method of interface MouseListener.
   * @param e MouseEvent
   */
  public void mouseClicked(MouseEvent e) {
    if ( (!running) && (btnStartStopResume.getText().equals("Start"))) { //works only if the simulation is not running and the first button is START
      isUntouched = false;
      int i = e.getX();
      int j = e.getY();
      int i0 = scEnergy.xToi(lastSelectedX);
      int j0 = scEnergy.yToj(lastSelectedE);

      //changing the initial direction of the arrow after double-click
      if ( (e.getClickCount() == 2) && (i0 - 8 < i) && (i0 + 8 > i) &&
          (j0 - 8 < j) && (j0 + 8 > j)) {
        ode.changeInitialDirection();
        ode.setInitialConditions(lastSelectedX, lastSelectedE);
        v = lastV * ode.getInitialDirection();
      }
      canvas.repaint();
    }
  }

  /**
   * Method of interface MouseListener.
   * @param e MouseEvent
   */
  public void mouseReleased(MouseEvent e) {
  }

  /**
   * Method of interface MouseListener.
   * @param e MouseEvent
   */
  public void mouseEntered(MouseEvent e) {
  }

  /**
   * Method of interface MouseListener.
   * @param e MouseEvent
   */
  public void mouseExited(MouseEvent e) {
  }

  /**
   * Method of interface MouseMotionListener.
   * @param e MouseEvent
   */
  public void mouseDragged(MouseEvent e) {
    isUntouched = false;

    int i = e.getX();
    int j = e.getY();

    //dragging the energy dot
    if ( (i < width - blankSpace) && (i > leftBlankSpace) &&
        (j < height - blankSpace) && (j > height - heightE) && editDot) { //if we are inside the energy diagram AND we are above the energy dot AND energy dot is editable.

      //this is done to be sure the energy dot is NOT dragged below the potential curve
      double x = scEnergy.iTox(i);
      double energy = scEnergy.jToy(j);
      if (U.evaluate(x) > energy)
        energy = U.evaluate(x);
      ode.setInitialConditions(x, energy);
      lastSelectedX = x;
      lastSelectedE = energy;
      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      v = lastV * ode.getInitialDirection();
      initialSetup();
      canvas.repaint();
    }

    //dragging the stationary points of potential
    else if ( (i < width - blankSpace) && (i > leftBlankSpace) &&
             (j < height - blankSpace) && (j > height - heightE) &&
             editPotential) { //if we are dragging inside the energy diagram AND potential is edited
      U.setHotPoint(draggedHotPointIndex, scEnergy.iTox(i), scEnergy.jToy(j));

      double x = scEnergy.iTox(i);
      double energy = scEnergy.jToy(j);
      if (U.evaluate(x) > energy)
        energy = U.evaluate(x);
      ode.setInitialConditions(lastSelectedX, lastSelectedE);

      lastV = Math.sqrt( (2.0 / 1.0) *
                        (lastSelectedE - U.evaluate(lastSelectedX)));
      v = lastV * ode.getInitialDirection();


      initialSetup();
      canvas.repaint();
    }

    //dragging the clockArrow
    else if ( (j < height - heightE - blankSpace) && editClockArrow) {
      double tA = scSpacetime.jToy(j);
      double xA = scSpacetime.iTox(i);
      double velocity = (xA - lastSelectedX) / tA;
      lastSelectedE = 0.5 * 1 * velocity * velocity + U.evaluate(lastSelectedX);
      if (lastSelectedE > 5)
        lastSelectedE = 5;

      if (velocity < 0) {
        ode.initialDirection = -1;
        lastV = Math.sqrt(2.0 / 1.0 * (lastSelectedE - U.evaluate(lastSelectedX)));
      }
      else {
        ode.initialDirection = 1;
        lastV = Math.sqrt(2.0 / 1.0 * (lastSelectedE - U.evaluate(lastSelectedX)));
      }

      initialSetup();
      canvas.repaint();
    }

    //now the cursor is set depending on the resulting value of "hand" variable
    if (editPotential || editDot || editClockArrow)
      canvas.setCursor(new Cursor(Cursor.HAND_CURSOR));
    else
      canvas.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));

  }

  /**
   * Method of interface MouseMotionListener.
   * @param e MouseEvent
   */
  public void mouseMoved(MouseEvent e) {
    if ( (!running) && (btnStartStopResume.getText().equals("Start"))) { //works only if the simulation is not running and the first button is START
      int i = e.getX();
      int j = e.getY();

      editPotential = false;
      editDot = false;
      editClockArrow = false;

      //Here it is determined if the potential is edited (the value editPotential is set)
      //and if so, the draggedHotPointIndex variable is determined
      Object hplist = U.getHotPointList();
      ListIterator hpit = ( (java.util.List) hplist).listIterator();
      int number = 0;
      while (hpit.hasNext()) {
        HotPoint sp = (HotPoint) hpit.next();
        number++;
        int ihp = scEnergy.xToi(sp.getX());
        int jhp = scEnergy.yToj(sp.getY());
        if ( (i < width - blankSpace) && (i > leftBlankSpace) &&
            (j < height - blankSpace) && (j > height - heightE) && (i < ihp + 8) &&
            (i > ihp - 8) && (j < jhp + 8) &&
            (j > jhp - 8)) { //if we are inside the energy diagram AND we are within 6 pixels off any of the stationary points
          editPotential = true;
          break;
        }
      }
      draggedHotPointIndex = number;

      //Here it is determined if the energy dot is edited (the value editDot is set)
      int i0 = scEnergy.xToi(ode.getX());
      int j0 = scEnergy.yToj(ode.getE());
      if ( (i < width - blankSpace) && (i > leftBlankSpace) &&
          (j < height - blankSpace) && (j > height - heightE) && (i0 - 8 < i) &&
          (i0 + 8 > i) && (j0 - 8 < j) && (j0 + 8 > j)) {
        editDot = true;
      }

      //Here it is determined if the clockArrow is edited (the value editClockArrow is set)
      if ( (j < height - heightE - blankSpace) && (clockHeadI - 16 < i) &&
          (clockHeadI + 16 > i) && (clockHeadJ - 16 < j) &&
          (clockHeadJ + 16 > j)) {
        editClockArrow = true;
      }

      //Now the cursor is set depending on the resulting value of "hand" variable
      if (editPotential || editDot || editClockArrow)
        canvas.setCursor(new Cursor(Cursor.HAND_CURSOR));
      else
        canvas.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));

    }

  }

}

/**
 * A canvas object for drawing.
 */
class Canvas
    extends JPanel {

  BufferedImage image;
  JEnergyDisplay applet;

  public Canvas(BufferedImage image, JEnergyDisplay applet) {
    this.image = image;
    this.applet = applet;
  }

  public void paintComponent(Graphics g) {
    Graphics2D g2 = (Graphics2D) g;
    applet.draw();
    g2.drawImage(image, null, 0, 0);
  }
}

/**
 * Records position and time of one event along a trajectory.
 */
class SolutionDatum {
  double x, v, t;

  public SolutionDatum(double x, double v, double t) {
    this.x = x;
    this.v = v;
    this.t = t;
  }

  public double getX() {
    return x;
  }

  public double getV() {
    return v;
  }

  public double getT() {
    return t;
  }
}