package org.opensourcephysics.tuleja.feynmanplate;

import org.opensourcephysics.numerics.*;
import org.opensourcephysics.displayejs.*;
import java.awt.Color;
import java.awt.Font;

/**
 * <p>Title: Disc</p>
 * <p>Description: Description: Subclass of RigidBody. It can draw itself and compute its motion.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * @author Alexander Tomori, Boris Gazovic, Slavomir Tuleja
 * @version 0.01
 */
public class Disc extends RigidBody{

  double radius=0.75;
  double h=0.05;//height
  InteractiveCylinder disc;
  InteractiveSurface medallion;
  int nRadialMed=3;
  int nAzimuthalMed=10;
  double[][][] dataMed=new double[nRadialMed][nAzimuthalMed][3];
  double[][][] dataMedTransformed=new double[nRadialMed][nAzimuthalMed][3];
  double[] customAxes=new double[9];

  /**
   * Constructor of the disc.
   */
  public Disc(){
    super(0.02982,0.05964);

    //disc
    disc=new InteractiveCylinder();
    disc.setXYZ(0.0, 0.0, -h/2.0);
    disc.setSizeXYZ(2.0*radius, 2*radius, h);
    disc.setResolution(new Resolution(5, 25));
    disc.getStyle().setEdgeColor(new Color(53, 97, 255));
    disc.setColorRegions(new double[]{}, new Color[]{new Color(198, 222, 255)});

    //medallion
    medallion=new InteractiveSurface();
    medallion.setColorRegions(new double[]{}, new Color[]{Color.red});
    medallion.getStyle().setEdgeColor(Color.RED);
    medallion.setDisplacementFactor(0.9);

    for(int i=0; i<nRadialMed; i++){
      for(int j=0; j<nAzimuthalMed; j++){
        double r=0.06 * i /(nRadialMed-1);
        double angle=2*Math.PI*j/(nAzimuthalMed-1);

        double x=r*Math.cos(angle)+0.9*radius;
        double y=r*Math.sin(angle);
        dataMed[i][j][0]=x;
        dataMed[i][j][1]=y;
        dataMed[i][j][2]=h/2.0;
      }
    }
    medallion.setData(dataMed);
    medallion.setOrigin(0,0,0,false);

    add(disc);
    add(medallion);
    setVisibility();
  }


  public void setVisibility(){
    super.setVisibilityExceptRigidBody();
    disc.setVisible(showRigidBody);
    medallion.setVisible(showRigidBody);
  }


  public void transformBodyOnly(double[][] rm){
    //transforms cilinder (disc)
    customAxes[0]=rm[0][0];
    customAxes[3]=rm[0][1];
    customAxes[6]=rm[0][2];
    customAxes[1]=rm[1][0];
    customAxes[4]=rm[1][1];
    customAxes[7]=rm[1][2];
    customAxes[2]=rm[2][0];
    customAxes[5]=rm[2][1];
    customAxes[8]=rm[2][2];

    disc.setCustomAxes(customAxes);
    double xBottomCenter, yBottomCenter, zBottomCenter;
    xBottomCenter=rm[0][0] * 0 + rm[0][1] * 0 + rm[0][2] * (-h/2.0);
    yBottomCenter=rm[1][0] * 0 + rm[1][1] * 0 + rm[1][2] * (-h/2.0);
    zBottomCenter=rm[2][0] * 0 + rm[2][1] * 0 + rm[2][2] * (-h/2.0);
    disc.setXYZ(xBottomCenter, yBottomCenter, zBottomCenter);

    //transforms medallion
    for (int i = 0; i < nRadialMed; i++) {
      for (int j = 0; j < nAzimuthalMed; j++) {
        double x = dataMed[i][j][0];
        double y = dataMed[i][j][1];
        double z = dataMed[i][j][2];

        double xn = rm[0][0] * x + rm[0][1] * y + rm[0][2] * z;
        double yn = rm[1][0] * x + rm[1][1] * y + rm[1][2] * z;
        double zn = rm[2][0] * x + rm[2][1] * y + rm[2][2] * z;

        dataMedTransformed[i][j][0] = xn;
        dataMedTransformed[i][j][1] = yn;
        dataMedTransformed[i][j][2] = zn;
      }
    }
    medallion.setData(dataMedTransformed);
  }
}
