package org.opensourcephysics.tuleja.feynmanplate;

import org.opensourcephysics.numerics.*;
import org.opensourcephysics.displayejs.*;
import java.awt.Color;
import java.awt.Font;

/**
 * <p>Title: Elementary plate</p>
 * <p>Description: Description: Subclass of RigidBody. It can draw itself and compute its motion.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * @author Alexander Tomori, Boris Gazovic, Slavomir Tuleja
 * @version 0.01
 */

import org.opensourcephysics.displayejs.*;

public class ElementaryPlate extends RigidBody {

  InteractiveSphere sph1, sph2, sph3, sph4;//
  InteractiveCylinder cil12, cil13, cil14, cil23, cil24, cil34;
  InteractiveArrow arr12, arr13, arr14, arr23, arr24, arr34;
  double[] data1=new double[3];
  double[] data2=new double[3];
  double[] data3=new double[3];
  double[] data4=new double[3];

  double r=0.75; //distance of each of four particles from the center of mass

  /**
   * Constructor of the elementary plate.
   */
  public ElementaryPlate(){
    super(0.02982,0.05964);

    sph1=new InteractiveSphere();
    sph2=new InteractiveSphere();
    sph3=new InteractiveSphere();
    sph4=new InteractiveSphere();

    sph1.getStyle().setFillPattern(new Color(53, 97, 255));
    sph1.getStyle().setEdgeColor(new Color(53, 97, 255));
    sph2.getStyle().setFillPattern(new Color(53, 97, 255));
    sph2.getStyle().setEdgeColor(new Color(53, 97, 255));
    sph3.getStyle().setFillPattern(new Color(53, 97, 255));
    sph3.getStyle().setEdgeColor(new Color(53, 97, 255));
    sph4.getStyle().setFillPattern(new Color(53, 97, 255));
    sph4.getStyle().setEdgeColor(new Color(53, 97, 255));

    double rSph=0.1;
    sph1.setSizeXYZ(rSph, rSph, rSph);
    sph2.setSizeXYZ(rSph, rSph, rSph);
    sph3.setSizeXYZ(rSph, rSph, rSph);
    sph4.setSizeXYZ(rSph, rSph, rSph);

    data1[0]=r;
    data1[1]=0.0;
    data1[2]=0.0;

    data2[0]=0.0;
    data2[1]=r;
    data2[2]=0.0;

    data3[0]=-r;
    data3[1]=0.0;
    data3[2]=0.0;

    data4[0]=0.0;
    data4[1]=-r;
    data4[2]=0.0;

    sph1.setXYZ(r, 0.0, 0.0);
    sph2.setXYZ(0.0,r, 0.0);
    sph3.setXYZ(-r, 0.0, 0.0);
    sph4.setXYZ(0.0, -r, 0.0);

    arr12=new InteractiveArrow(InteractiveArrow.SEGMENT);
    arr13=new InteractiveArrow(InteractiveArrow.SEGMENT);
    arr14=new InteractiveArrow(InteractiveArrow.SEGMENT);
    arr23=new InteractiveArrow(InteractiveArrow.SEGMENT);
    arr24=new InteractiveArrow(InteractiveArrow.SEGMENT);
    arr34=new InteractiveArrow(InteractiveArrow.SEGMENT);

    arr12.getStyle().setEdgeColor(new Color(53, 97, 255));
    arr13.getStyle().setEdgeColor(new Color(53, 97, 255));
    arr14.getStyle().setEdgeColor(new Color(53, 97, 255));
    arr23.getStyle().setEdgeColor(new Color(53, 97, 255));
    arr24.getStyle().setEdgeColor(new Color(53, 97, 255));
    arr34.getStyle().setEdgeColor(new Color(53, 97, 255));

    setVisibility();

    add(sph1);
    add(sph2);
    add(sph3);
    add(sph4);

    add(arr12);
    add(arr13);
    add(arr14);
    add(arr23);
    add(arr24);
    add(arr34);
  }


  public void setVisibility(){
    super.setVisibilityExceptRigidBody();
    arr12.setVisible(showRigidBody);
    arr13.setVisible(showRigidBody);
    arr14.setVisible(showRigidBody);
    arr23.setVisible(showRigidBody);
    arr24.setVisible(showRigidBody);
    arr34.setVisible(showRigidBody);
    sph1.setVisible(showRigidBody);
    sph2.setVisible(showRigidBody);
    sph3.setVisible(showRigidBody);
    sph4.setVisible(showRigidBody);
  }


  public void transformBodyOnly(double[][] rm){
    double[] data;

    data=transformPointsOnly(data1, rm);
    sph1.setXYZ(data[0], data[1], data[2]);

    data=transformPointsOnly(data2, rm);
    sph2.setXYZ(data[0], data[1], data[2]);

    data=transformPointsOnly(data3, rm);
    sph3.setXYZ(data[0], data[1], data[2]);

    data=transformPointsOnly(data4, rm);
    sph4.setXYZ(data[0], data[1], data[2]);

    data=transformPointsOnly(data4, rm);
    arr12.setXYZ(data[0], data[1], data[2]);

    double[] dataA, dataB;//starting and ending point

    dataA=transformPointsOnly(data1, rm);
    dataB=transformPointsOnly(data2, rm);
    arr12.setXYZ(dataA[0], dataA[1], dataA[2]);
    arr12.setSizeXYZ(dataB[0]-dataA[0], dataB[1]-dataA[1], dataB[2]-dataA[2]);

    dataA=transformPointsOnly(data1, rm);
    dataB=transformPointsOnly(data3, rm);
    arr13.setXYZ(dataA[0], dataA[1], dataA[2]);
    arr13.setSizeXYZ(dataB[0]-dataA[0], dataB[1]-dataA[1], dataB[2]-dataA[2]);

    dataA=transformPointsOnly(data1, rm);
    dataB=transformPointsOnly(data4, rm);
    arr14.setXYZ(dataA[0], dataA[1], dataA[2]);
    arr14.setSizeXYZ(dataB[0]-dataA[0], dataB[1]-dataA[1], dataB[2]-dataA[2]);

    dataA=transformPointsOnly(data2, rm);
    dataB=transformPointsOnly(data3, rm);
    arr23.setXYZ(dataA[0], dataA[1], dataA[2]);
    arr23.setSizeXYZ(dataB[0]-dataA[0], dataB[1]-dataA[1], dataB[2]-dataA[2]);

    dataA=transformPointsOnly(data2, rm);
    dataB=transformPointsOnly(data4, rm);
    arr24.setXYZ(dataA[0], dataA[1], dataA[2]);
    arr24.setSizeXYZ(dataB[0]-dataA[0], dataB[1]-dataA[1], dataB[2]-dataA[2]);

    dataA=transformPointsOnly(data3, rm);
    dataB=transformPointsOnly(data4, rm);
    arr34.setXYZ(dataA[0], dataA[1], dataA[2]);
    arr34.setSizeXYZ(dataB[0]-dataA[0], dataB[1]-dataA[1], dataB[2]-dataA[2]);
  }

  public double[] transformPointsOnly(double[] data, double[][] rm){
    double[] dataRotated=new double[3];

    dataRotated[0]= data[0]*rm[0][0] + data[1]*rm[0][1] + data[2]*rm[0][2];
    dataRotated[1]= data[0]*rm[1][0] + data[1]*rm[1][1] + data[2]*rm[1][2];
    dataRotated[2]= data[0]*rm[2][0] + data[1]*rm[2][1] + data[2]*rm[2][2];

    return dataRotated;
  }


}
