package org.opensourcephysics.tuleja.jrelativisticparadoxes;

/**
 * <p>Title: JRelativisticParadoxes</p>
 * <p>Description: Animates paradoxes of Special Relativity from within different frames.</p>
 * <p>Copyright: Copyright (c) 2003 Slavomir Tuleja</p>
 * @author Slavomir Tuleja
 * @version 1.0
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;

public class JBulbParadox
    extends JApplet
    implements ActionListener, Runnable, MouseListener, MouseMotionListener {
  private JButton btnStartPause, btnReset, btnStepForward, btnStepBackward;
  private JLabel lblVelocity;
  private JTextField tfVelocity;
  private JComboBox comboFrame;
  private ComboHandler comboHandler;
  private int w = 451, h = 301;
  private BufferedImage image;
  private Graphics2D g2;
  private JPanel panelControls;
  private Canvas canvas;
  private Container cp;
  private Thread thread = null;
  private final int SLEEP_TIME = 50;
  private Scale sc;
  boolean stop=true, laboratory = true, cursorOver=false, isDragged=false;
  double t=0, dt=0.02;
  BasicStroke bsThick, bsThin;

  double a,b,c,d,gama,v, q=1.5;
  double xOld, x, xCur;

  public void init() {
    image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
    canvas = new Canvas(image, this);
    btnStartPause = new JButton("Start");
    btnStartPause.setToolTipText("Starts, pauses or resumes the animation.");
    btnReset = new JButton("Reset");
    btnReset.setToolTipText("Reads velocity and resets the animation.");
    btnStepForward = new JButton(">>");
    btnStepForward.setToolTipText("Step forward.");
    btnStepBackward = new JButton("<<");
    btnStepBackward.setToolTipText("Step back.");
    lblVelocity = new JLabel("Velocity: ");
    tfVelocity = new JTextField("0.866");
    tfVelocity.setToolTipText("Relative velocity of a shuttle and the rails.");
    String[] items = {"Laboratory","Rocket"};
    comboFrame = new JComboBox(items);
    comboFrame.setToolTipText("Please choose the frame.");

    panelControls = new JPanel();
    panelControls.setLayout(new BoxLayout(panelControls,BoxLayout.X_AXIS));
    panelControls.add(btnStartPause);
    panelControls.add(btnStepBackward);
    panelControls.add(btnStepForward);
    panelControls.add(btnReset);
    panelControls.add(Box.createRigidArea(new Dimension(5,0)));
    panelControls.add(lblVelocity);
    panelControls.add(tfVelocity);
    panelControls.add(comboFrame);

    cp = getContentPane();
    cp.setLayout(new BorderLayout());
    cp.add(canvas, BorderLayout.CENTER);
    cp.add(panelControls, BorderLayout.SOUTH);

    btnStartPause.addActionListener(this);
    btnStepBackward.addActionListener(this);
    btnStepForward.addActionListener(this);
    btnReset.addActionListener(this);
    comboHandler = new ComboHandler(this);
    comboFrame.addActionListener(comboHandler);
    canvas.addMouseListener(this);
    canvas.addMouseMotionListener(this);

    g2 = image.createGraphics();
    g2.setRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING,
                                              RenderingHints.VALUE_ANTIALIAS_ON));

    g2.setColor(Color.white);
    g2.fillRect(0,0,w,h);
    g2.setColor(Color.black);
    g2.drawRect(0,0,w-1,h-1);

    bsThick = new BasicStroke(2.0f,BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND);
    bsThin = new BasicStroke(1.0f,BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND);
    setFrame();
    cp.repaint();
  }

  public void actionPerformed(ActionEvent event) {
    if(event.getActionCommand().equals("Start")){
      t=0;
      btnStartPause.setText("Pause");
      stop=false;
      thread = new Thread(this);
      thread.start();
    }
    else if(event.getActionCommand().equals("Pause")){
      btnStartPause.setText("Resume");
      stop=true;
    }
    else if(event.getActionCommand().equals("Resume")){
      btnStartPause.setText("Pause");
      stop=false;
      thread = new Thread(this);
      thread.start();
    }
    else if(event.getActionCommand().equals("Reset")){
      btnStartPause.setText("Start");
      setFrame();
      t=0;
      stop=true;
      canvas.repaint();
    }
    else if(event.getActionCommand().equals("<<")){
      t=t-dt;
      canvas.repaint();
    }
    else if(event.getActionCommand().equals(">>")){
      t=t+dt;
      canvas.repaint();
    }
  }



  public void run() {
    do{
      t+=dt;
      canvas.repaint();
      try {
        thread.sleep(SLEEP_TIME);
      }
      catch (InterruptedException exception) {
        System.err.println("Interrupted");
      }
    }
    while (!stop);
  }

  public void draw(Graphics2D g2) {
    if(laboratory) g2.drawString("Laboratory frame plot",10,15);
      else g2.drawString("Rocket frame plot",10,15);
    g2.drawString("time = ".concat(MyUtils.roundOff(t,2)).concat(" m"),10,30);

    if(laboratory) drawLaboratory(g2);
      else drawRocket(g2);

    g2.setColor(Color.black);
    if(cursorOver && !isDragged) {
      if(laboratory) g2.drawString("x = ".concat(MyUtils.roundOff(xCur,2)).concat(" m"),10,290);
        else g2.drawString("x' = ".concat(MyUtils.roundOff(xCur,2)).concat(" m"),10,290);
    }
  }

  public void drawLaboratory(Graphics2D g2){
    double l1=7;
    double l2=0.5;
    double l3=0.1;
    double r=0.2;

    g2.setStroke(bsThick);

    g2.setColor(Color.blue);
    //upper rail
    g2.drawLine(sc.xToi(-l1),sc.yToj(0.75),sc.xToi(0),sc.yToj(0.75));
    g2.drawLine(sc.xToi(0),sc.yToj(0.75),sc.xToi(0),sc.yToj(1.05));
    g2.drawLine(sc.xToi(0),sc.yToj(1.05),sc.xToi(2),sc.yToj(1.05));
    g2.drawLine(sc.xToi(2),sc.yToj(1.05),sc.xToi(2),sc.yToj(0.75));
    g2.drawLine(sc.xToi(2),sc.yToj(0.75),w,sc.yToj(0.75));

    //lower rail
    g2.drawLine(sc.xToi(-l1),sc.yToj(-0.75),w,sc.yToj(-0.75));

    double bulbPos=l1+l3+r;
    double t1=bulbPos;
    double t2=bulbPos+2/(gama*v);
    double t3=bulbPos+2+2/v;
    double t4=bulbPos+2+(2+(2/gama))/v;

    //bulb
    if(t>t4) g2.setColor(Color.yellow);
      else if(t>t3) g2.setColor(Color.red);
        else if(t>t2) g2.setColor(Color.black);
          else if(t>t1) g2.setColor(Color.red);
            else g2.setColor(Color.yellow);
    g2.fillOval(sc.xToi(-l1-l3-2*r), sc.yToj(0.75+l2+r),sc.xToi(-l1-l3)-sc.xToi(-l1-l3-2*r),sc.yToj(0.75+l2-r)-sc.yToj(0.75+l2+r));
    g2.setColor(Color.blue);
    g2.drawLine(sc.xToi(-l1),sc.yToj(0.75),sc.xToi(-l1),sc.yToj(0.75+l2));
    g2.drawLine(sc.xToi(-l1),sc.yToj(0.75+l2),sc.xToi(-l1-l3),sc.yToj(0.75+l2));
    g2.drawLine(sc.xToi(-l1-l3-2*r),sc.yToj(0.75+l2),sc.xToi(-l1-l3-2*r-l3),sc.yToj(0.75+l2));
    g2.drawLine(sc.xToi(-l1-l3-2*r-l3),sc.yToj(0.75+l2),sc.xToi(-l1-l3-2*r-l3),sc.yToj(0.05));
    g2.drawLine(sc.xToi(-l1-l3-2*r-l3),sc.yToj(-0.05),sc.xToi(-l1-l3-2*r-l3),sc.yToj(-0.75));
    g2.drawLine(sc.xToi(-l1-l3-2*r-l3),sc.yToj(-0.75),sc.xToi(-l1),sc.yToj(-0.75));
    g2.drawLine(sc.xToi(-l1-l3-2*r-l3-0.15),sc.yToj(0.05),sc.xToi(-l1-l3-2*r-l3+0.15),sc.yToj(0.05));
    g2.drawLine(sc.xToi(-l1-l3-2*r-l3-0.07),sc.yToj(-0.05),sc.xToi(-l1-l3-2*r-l3+0.07),sc.yToj(-0.05));
    g2.drawOval(sc.xToi(-l1-l3-2*r), sc.yToj(0.75+l2+r),sc.xToi(-l1-l3)-sc.xToi(-l1-l3-2*r),sc.yToj(0.75+l2-r)-sc.yToj(0.75+l2+r));
    double x1=-l1-l3-r-r*Math.cos(Math.PI/4.0);
    double y1=0.75+l2+r*Math.sin(Math.PI/4.0);
    double x2=-l1-l3-r+r*Math.cos(Math.PI/4.0);
    double y2=0.75+l2-r*Math.sin(Math.PI/4.0);
    g2.drawLine(sc.xToi(x1),sc.yToj(y1),sc.xToi(x2),sc.yToj(y2));
    g2.drawLine(sc.xToi(x2),sc.yToj(y1),sc.xToi(x1),sc.yToj(y2));




    //horizontal insulators
    g2.setColor(Color.lightGray);
    g2.fillRect(sc.xToi(v*t-2/gama),sc.yToj(0.5),sc.xToi(v*t)-sc.xToi(v*t-2/gama),sc.yToj(-0.5)-sc.yToj(0.5));
    g2.setColor(Color.gray);
    g2.drawLine(sc.xToi(v*t-2/gama),sc.yToj(0.5),sc.xToi(v*t),sc.yToj(0.5));
    g2.drawLine(sc.xToi(v*t-2/gama),sc.yToj(-0.5),sc.xToi(v*t),sc.yToj(-0.5));

    //vertical conductor A
    g2.setColor(Color.blue);
    g2.drawLine(sc.xToi(v*t-2/gama),sc.yToj(0.75),sc.xToi(v*t-2/gama),sc.yToj(-0.75));
    g2.drawString("A",sc.xToi(v*t-2/gama)-13,sc.yToj(0)-5);

    //vertical conductor B
    g2.drawLine(sc.xToi(v*t),sc.yToj(0.75),sc.xToi(v*t),sc.yToj(-0.75));
    g2.drawString("B",sc.xToi(v*t)+5,sc.yToj(0)-5);

    //information "disconnected at B"
    g2.setStroke(bsThin);
    g2.setColor(Color.black);
    if(t>0){
      g2.drawLine(sc.xToi(-t),sc.yToj(1.3),sc.xToi(-t),sc.yToj(-0.8)+60);
      g2.drawString("disconnected at B",sc.xToi(-t)+5,sc.yToj(-0.8)+60);
    }

    //information "disconnected at A"
    g2.setColor(Color.black);
    if(t>=2/(gama*v)){
      g2.drawLine(sc.xToi(-(t-2/(gama*v))),sc.yToj(1.3),sc.xToi(-(t-2/(gama*v))),sc.yToj(-0.8)+45);
      g2.drawString("disconnected at A",sc.xToi(-(t-2/(gama*v)))+5,sc.yToj(-0.8)+45);
    }

    //information "connected at B"
    g2.setColor(Color.red);
    if(t>=2/v){
      g2.drawLine(sc.xToi(-(t-2/v)+2),sc.yToj(1.3),sc.xToi(-(t-2/v)+2),sc.yToj(-0.8)+30);
      g2.drawString("connected at B",sc.xToi(-(t-2/v)+2)+5,sc.yToj(-0.8)+30);
    }

    //information "connected at A"
    g2.setColor(Color.red);
    if(t>=(2+(2/gama))/v){
      g2.drawLine(sc.xToi(-(t-(2+2/gama)/v)+2),sc.yToj(1.3),sc.xToi(-(t-(2+2/gama)/v)+2),sc.yToj(-0.8)+15);
      g2.drawString("connected at A",sc.xToi(-(t-(2+2/gama)/v)+2)+5,sc.yToj(-0.8)+15);
    }
  }


  public void drawRocket(Graphics2D g2){
    double l1=7;
    double l2=0.5;
    double l3=0.1;
    double r=0.2;

    g2.setStroke(bsThick);
    g2.setColor(Color.blue);

    //upper rail
    g2.drawLine(sc.xToi(-l1/gama-v*t),sc.yToj(0.75),sc.xToi(-v*t),sc.yToj(0.75));
    g2.drawLine(sc.xToi(-v*t),sc.yToj(0.75),sc.xToi(-v*t),sc.yToj(1.05));
    g2.drawLine(sc.xToi(-v*t),sc.yToj(1.05),sc.xToi(-v*t+2/gama),sc.yToj(1.05));
    g2.drawLine(sc.xToi(-v*t+2/gama),sc.yToj(1.05),sc.xToi(-v*t+2/gama),sc.yToj(0.75));
    g2.drawLine(sc.xToi(-v*t+2/gama),sc.yToj(0.75),w,sc.yToj(0.75));

    //lower rail
    g2.drawLine(sc.xToi(-l1/gama-v*t),sc.yToj(-0.75),w,sc.yToj(-0.75));

    //bulb
    double bulbPos=(l1+l3+r);
    double t1=bulbPos/(gama*(1-v));
    double t2=bulbPos/(gama*(1-v))+2/v;
    double t3=bulbPos/(gama*(1-v))+2/gama * 1/(v*(1-v));
    double t4=bulbPos/(gama*(1-v))+2/v+ 2/gama * 1/(v*(1-v));

    if(t>t4) g2.setColor(Color.yellow);
      else if(t>t3) g2.setColor(Color.red);
        else if(t>t2) g2.setColor(Color.black);
          else if(t>t1) g2.setColor(Color.red);
            else g2.setColor(Color.yellow);
    g2.fillOval(sc.xToi((-l1-l3-2*r)/gama - v*t), sc.yToj(0.75+l2+r),sc.xToi((-l1-l3)/gama - v*t)-sc.xToi((-l1-l3-2*r)/gama - v*t),sc.yToj(0.75+l2-r)-sc.yToj(0.75+l2+r));
    g2.setColor(Color.blue);
    g2.drawLine(sc.xToi(-l1/gama - v*t),sc.yToj(0.75),sc.xToi(-l1/gama - v*t),sc.yToj(0.75+l2));
    g2.drawLine(sc.xToi(-l1/gama - v*t),sc.yToj(0.75+l2),sc.xToi((-l1-l3)/gama - v*t),sc.yToj(0.75+l2));
    g2.drawLine(sc.xToi((-l1-l3-2*r)/gama - v*t),sc.yToj(0.75+l2),sc.xToi((-l1-l3-2*r-l3)/gama - v*t),sc.yToj(0.75+l2));
    g2.drawLine(sc.xToi((-l1-l3-2*r-l3)/gama - v*t),sc.yToj(0.75+l2),sc.xToi((-l1-l3-2*r-l3)/gama - v*t),sc.yToj(0.05));
    g2.drawLine(sc.xToi((-l1-l3-2*r-l3)/gama - v*t),sc.yToj(-0.05),sc.xToi((-l1-l3-2*r-l3)/gama - v*t),sc.yToj(-0.75));
    g2.drawLine(sc.xToi((-l1-l3-2*r-l3)/gama - v*t),sc.yToj(-0.75),sc.xToi(-l1/gama - v*t),sc.yToj(-0.75));
    g2.drawLine(sc.xToi((-l1-l3-2*r-l3-0.15)/gama - v*t),sc.yToj(0.05),sc.xToi((-l1-l3-2*r-l3+0.15)/gama - v*t),sc.yToj(0.05));
    g2.drawLine(sc.xToi((-l1-l3-2*r-l3-0.07)/gama - v*t),sc.yToj(-0.05),sc.xToi((-l1-l3-2*r-l3+0.07)/gama - v*t),sc.yToj(-0.05));
    g2.drawOval(sc.xToi((-l1-l3-2*r)/gama - v*t), sc.yToj(0.75+l2+r),sc.xToi((-l1-l3)/gama - v*t)-sc.xToi((-l1-l3-2*r)/gama - v*t),sc.yToj(0.75+l2-r)-sc.yToj(0.75+l2+r));
    double x1=(-l1-l3-r-r*Math.cos(Math.PI/4.0))/gama - v*t;
    double y1=0.75+l2+r*Math.sin(Math.PI/4.0);
    double x2=(-l1-l3-r+r*Math.cos(Math.PI/4.0))/gama - v*t;
    double y2=0.75+l2-r*Math.sin(Math.PI/4.0);
    g2.drawLine(sc.xToi(x1),sc.yToj(y1),sc.xToi(x2),sc.yToj(y2));
    g2.drawLine(sc.xToi(x2),sc.yToj(y1),sc.xToi(x1),sc.yToj(y2));


    //horizontal insulators
    g2.setColor(Color.lightGray);
    g2.fillRect(sc.xToi(-2),sc.yToj(0.5),sc.xToi(0)-sc.xToi(-2),sc.yToj(-0.5)-sc.yToj(0.5));
    g2.setColor(Color.gray);
    g2.drawLine(sc.xToi(-2),sc.yToj(0.5),sc.xToi(0),sc.yToj(0.5));
    g2.drawLine(sc.xToi(-2),sc.yToj(-0.5),sc.xToi(0),sc.yToj(-0.5));

    //vertical conductor A
    g2.setColor(Color.blue);
    g2.drawLine(sc.xToi(-2),sc.yToj(0.75),sc.xToi(-2),sc.yToj(-0.75));
    g2.drawString("A",sc.xToi(-2)-13,sc.yToj(0)-5);

    //vertical conductor B
    g2.drawLine(sc.xToi(0),sc.yToj(0.75),sc.xToi(0),sc.yToj(-0.75));
    g2.drawString("B",sc.xToi(0)+5,sc.yToj(0)-5);

    //information "disconnected at B"
    g2.setStroke(bsThin);
    g2.setColor(Color.black);
    if(t>0){
      g2.drawLine(sc.xToi(-t),sc.yToj(1.3),sc.xToi(-t),sc.yToj(-0.8)+60);
      g2.drawString("disconnected at B",sc.xToi(-t)+5,sc.yToj(-0.8)+60);
    }

    //information "connected at B"
    g2.setColor(Color.red);
    if(t>=2/(gama*v)){
      g2.drawLine(sc.xToi(-(t-2/(gama*v))),sc.yToj(1.3),sc.xToi(-(t-2/(gama*v))),sc.yToj(-0.8)+45);
      g2.drawString("connected at B",sc.xToi(-(t-2/(gama*v)))+5,sc.yToj(-0.8)+45);
    }

    //information "disconnected at A"
    g2.setColor(Color.black);
    if(t>=2/v){
      g2.drawLine(sc.xToi(-(t-2/v)-2),sc.yToj(1.3),sc.xToi(-(t-2/v)-2),sc.yToj(-0.8)+30);
      g2.drawString("disconnected at A",sc.xToi(-(t-2/v)-2)+5,sc.yToj(-0.8)+30);
    }

    //information "connected at A"
    g2.setColor(Color.red);
    if(t>=(2+(2/gama))/v){
      g2.drawLine(sc.xToi(-(t-(2+2/gama)/v)-2),sc.yToj(1.3),sc.xToi(-(t-(2+2/gama)/v)-2),sc.yToj(-0.8)+15);
      g2.drawString("connected at A",sc.xToi(-(t-(2+2/gama)/v)-2)+5,sc.yToj(-0.8)+15);
    }
  }

  public String adjustdot(String s) {
    return s.replace(',', '.');
  }

  public void setVelocity(){
    String svOld = String.valueOf(v);
    double vOld = v;
    try {
      v = Double.valueOf(adjustdot(tfVelocity.getText())).doubleValue();
    }
    catch (NumberFormatException ex) {
      JOptionPane.showMessageDialog(this,
          "Wrong input format!\nPlease, try it again.",
          "Input error", JOptionPane.ERROR_MESSAGE);
      tfVelocity.setText(svOld);
    }

    if(!((v>=0)&&(v<=0.95))) {
      JOptionPane.showMessageDialog(this,
          "Your input value should be between 0 and 0.95!\nPlease, try it again.",
          "Input range error", JOptionPane.ERROR_MESSAGE);
      tfVelocity.setText(svOld);
      v=vOld;
    }
    gama = 1.0 / Math.sqrt(1 - v * v);
  }

  public void setFrame(){
    btnStartPause.setText("Start");
    stop=true;
    setVelocity();

    int index = comboFrame.getSelectedIndex();
    if(index==0) {//Laboratory frame
      laboratory=true;
    }
    else {//Rocket frame
      laboratory=false;
    }

    a=-5;
    b=3;
    c=-2;
    d=2;
    sc = new Scale(a, b ,c ,d , 0, 450, 0, 300);
    t=0;
    canvas.repaint();
  }

  public void mouseClicked(MouseEvent event){
  }

  public void mouseEntered(MouseEvent event){
    canvas.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
  }

  public void mouseExited(MouseEvent event){
    cursorOver=false;
    canvas.repaint();
    canvas.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
  }

  public void mousePressed(MouseEvent event){
    isDragged=true;
    canvas.setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
    xOld=sc.iTox(event.getX());
  }

  public void mouseReleased(MouseEvent event){
    isDragged=false;
    canvas.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
  }

  public void mouseDragged(MouseEvent event){
    x=sc.iTox(event.getX());
    a=a-1*(x-xOld);
    b=b-1*(x-xOld);
    sc = new Scale(a, b ,c ,d , 0, 450, 0, 300);
    canvas.repaint();
  }

  public void mouseMoved(MouseEvent event){
    cursorOver=true;
    xCur=sc.iTox(event.getX());
    canvas.repaint();
  }



  public String getAppletInfo() {
    return "(c) 2003 Slavomir Tuleja";
  }

  public static void main(String[] args){
    JFrame fr = new JFrame("JBulbParadox");
    fr.setSize(459,362);
    JBulbParadox applet = new JBulbParadox();
    applet.init();
    applet.start();
    fr.getContentPane().add(applet,BorderLayout.CENTER);
    fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    fr.setVisible(true);
  }

}

class Canvas
    extends JPanel {
  BufferedImage image;
  JBulbParadox applet;

  public Canvas(BufferedImage image, JBulbParadox applet) {
    this.image = image;
    this.applet = applet;
  }

  public void paintComponent(Graphics g) {
    Graphics2D g2 = (Graphics2D) g;
    g2.setRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING,
                                              RenderingHints.VALUE_ANTIALIAS_ON));
    g2.drawImage(image, null, 0, 0);
    applet.draw(g2);
  }

}

class ComboHandler implements ActionListener {
   private JBulbParadox applet;

   public ComboHandler(JBulbParadox applet){
      this.applet=applet;
   }

   public void actionPerformed(ActionEvent e){
      applet.setFrame();
   }
}
