package org.opensourcephysics.tuleja.orbits;

public class Orbit implements OrbitODE {
   private double[] state= new double[] {0.0,0.0,0.0,0.0,0.0,0.0,0.0};
   //...initial state {x,y,z,vx,vy,vz,t}
   private double GM=3.989e14;

   public double[] getState(){
      return state;
   }

   public void getRate(double[] state, double[] rates){
      double rrr=Math.pow(state[0]*state[0]+state[1]*state[1]+state[2]*state[2],1.5);
      rates[0]=state[3];
      rates[1]=state[4];
      rates[2]=state[5];
      rates[3]=-GM*state[0]/rrr;
      rates[4]=-GM*state[1]/rrr;
      rates[5]=-GM*state[2]/rrr;
      rates[6]=1;
   }

   public void setIC(double r, double theta, double phi, double vR, double vTheta, double vPhi) {
      state[0]=r*Math.sin(theta)*Math.cos(phi);
      state[1]=r*Math.sin(theta)*Math.sin(phi);
      state[2]=r*Math.cos(theta);
      state[3]=vR*Math.sin(theta)*Math.cos(phi)-vPhi*Math.sin(phi)+vTheta*Math.cos(theta)*Math.cos(phi);
      state[4]=vR*Math.sin(theta)*Math.sin(phi)+vPhi*Math.cos(phi)+vTheta*Math.cos(theta)*Math.sin(phi);
      state[5]=vR*Math.cos(theta)-vTheta*Math.sin(theta);
      state[6]=0;
   }

   public void setT(double t){
      state[6]=t;
   }

   public double getR(){
      return Math.sqrt(state[0]*state[0]+state[1]*state[1]+state[2]*state[2]);
   }

   public double getPhi(){
      return Math.atan2(state[1],state[0]);
   }

   public double getTheta(){
      return Math.acos(state[2]/Math.sqrt(state[0]*state[0]+state[1]*state[1]+state[2]*state[2]));
   }

   public double getVR(){
      double vx=state[3];
      double vy=state[4];
      double vz=state[5];
      return vx*Math.cos(getPhi())*Math.sin(getTheta()) + vy*Math.sin(getPhi())*Math.sin(getTheta()) + vz*Math.cos(getTheta());
   }

   public double getVTheta(){
      double vx=state[3];
      double vy=state[4];
      double vz=state[5];
      return vx*Math.cos(getPhi())*Math.cos(getTheta()) + vy*Math.sin(getPhi())*Math.cos(getTheta()) - vz*Math.sin(getTheta());
   }

   public double getVPhi(){
      double vx=state[3];
      double vy=state[4];
      double vz=state[5];
      return vy*Math.cos(getPhi()) - vx*Math.sin(getPhi());
   }

   public double getT(){
      return state[6];
   }

   public double getX(){
      return state[0];
   }

   public double getY(){
      return state[1];
   }

   public double getZ(){
      return state[2];
   }
}
